"""The transport decorator (~transport).

This decorator allows changes to agent response behaviour and queue status updates.
"""

from typing import Optional

from marshmallow import EXCLUDE, fields, validate

from ..models.base import BaseModel, BaseModelSchema
from ..valid import UUID4_EXAMPLE, WHOLE_NUM_EXAMPLE, WHOLE_NUM_VALIDATE


class TransportDecorator(BaseModel):
    """Class representing the transport decorator."""

    class Meta:
        """TransportDecorator metadata."""

        schema_class = "TransportDecoratorSchema"

    def __init__(
        self,
        *,
        return_route: Optional[str] = None,
        return_route_thread: Optional[str] = None,
        queued_message_count: Optional[int] = None,
    ):
        """Initialize a TransportDecorator instance.

        Args:
            return_route: Set the return routing mode
            return_route_thread: Identify the thread to enable return routing for
            queued_message_count: Indicate the number of queued messages
        """
        super().__init__()
        self.return_route = return_route
        self.return_route_thread = return_route_thread
        self.queued_message_count = queued_message_count


class TransportDecoratorSchema(BaseModelSchema):
    """Transport decorator schema used in serialization/deserialization."""

    class Meta:
        """TransportDecoratorSchema metadata."""

        model_class = TransportDecorator
        unknown = EXCLUDE

    return_route = fields.Str(
        required=False,
        validate=validate.OneOf(["none", "all", "thread"]),
        metadata={
            "description": "Return routing mode: none, all, or thread",
            "example": "all",
        },
    )
    return_route_thread = fields.Str(
        required=False,
        metadata={
            "description": "Identifier for thread for which to enable return routing",
            "example": UUID4_EXAMPLE,
        },
    )
    queued_message_count = fields.Int(
        required=False,
        validate=WHOLE_NUM_VALIDATE,
        metadata={
            "description": "Number of queued messages",
            "strict": True,
            "example": WHOLE_NUM_EXAMPLE,
        },
    )



"""RouteManager provider."""

from ....config.base import BaseInjector, BaseProvider, BaseSettings
from ....core.profile import Profile
from ....multitenant.base import BaseMultitenantManager
from ....multitenant.route_manager import BaseWalletRouteManager, MultitenantRouteManager
from .route_manager import CoordinateMediationV1RouteManager


class RouteManagerProvider(BaseProvider):
    """Route manager provider.

    Decides which route manager to use based on settings.
    """

    def __init__(self, root_profile: Profile):
        """Initialize route manager provider."""
        self.root_profile = root_profile

    def provide(self, settings: BaseSettings, injector: BaseInjector):
        """Create the appropriate route manager instance."""
        wallet_id = settings.get("wallet.id")
        multitenant_mgr = self.root_profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            if wallet_id:
                return MultitenantRouteManager(self.root_profile)
            else:
                return BaseWalletRouteManager()

        return CoordinateMediationV1RouteManager()



import pytest
import pytest_asyncio

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......utils.testing import create_test_profile
from ...messages.hangup import Hangup
from .. import hangup_handler as test_module


@pytest_asyncio.fixture
async def request_context():
    ctx = RequestContext.test_context(await create_test_profile())
    yield ctx


class TestHangupHandler:
    """Unit tests for HangupHandler."""

    @pytest.mark.asyncio
    @mock.patch.object(test_module, "DIDRotateManager")
    async def test_handle(self, MockDIDRotateManager, request_context):
        MockDIDRotateManager.return_value.receive_hangup = mock.CoroutineMock()

        request_context.message = Hangup()
        request_context.connection_record = mock.MagicMock()
        request_context.connection_ready = True

        handler = test_module.HangupHandler()
        responder = MockResponder()
        await handler.handle(request_context, responder)

        MockDIDRotateManager.return_value.receive_hangup.assert_called_once_with(
            request_context.connection_record
        )



"""Test IndyDIDResolver."""

import pytest
import pytest_asyncio

from ....core.profile import Profile
from ....ledger.base import BaseLedger
from ....ledger.error import LedgerError
from ....ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ....messaging.valid import IndyDID
from ....multitenant.base import BaseMultitenantManager
from ....multitenant.manager import MultitenantManager
from ....tests import mock
from ....utils.testing import create_test_profile
from ...base import DIDNotFound, ResolverError
from ..indy import IndyDIDResolver, _routing_keys_as_did_key_urls

# pylint: disable=W0621
TEST_DID0 = "did:sov:WgWxqztrNooG92RXvxSTWv"


@pytest.fixture
def resolver():
    """Resolver fixture."""
    yield IndyDIDResolver()


@pytest.fixture
def ledger():
    """Ledger fixture."""
    ledger = mock.MagicMock(spec=BaseLedger)
    ledger.get_all_endpoints_for_did = mock.CoroutineMock(
        return_value={
            "endpoint": "https://github.com/",
            "profile": "https://example.com/profile",
        }
    )
    ledger.get_key_for_did = mock.CoroutineMock(return_value="key")
    yield ledger


@pytest_asyncio.fixture
async def profile(ledger):
    """Profile fixture."""
    profile = await create_test_profile()
    mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
    mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
        return_value=(None, ledger)
    )
    profile.context.injector.bind_instance(IndyLedgerRequestsExecutor, mock_executor)
    yield profile


class TestIndyResolver:
    @pytest.mark.asyncio
    async def test_supported_did_regex(self, profile, resolver: IndyDIDResolver):
        """Test the supported_did_regex."""
        assert resolver.supported_did_regex == IndyDID.PATTERN
        assert await resolver.supports(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_resolve(self, profile: Profile, resolver: IndyDIDResolver):
        """Test resolve method."""
        assert await resolver.resolve(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_resolve_with_accept(self, profile: Profile, resolver: IndyDIDResolver):
        """Test resolve method."""
        assert await resolver.resolve(
            profile, TEST_DID0, ["didcomm/aip1", "didcomm/aip2;env=rfc19"]
        )

    @pytest.mark.asyncio
    async def test_resolve_multitenant(
        self, profile: Profile, resolver: IndyDIDResolver, ledger: BaseLedger
    ):
        """Test resolve method."""
        profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=("test_ledger_id", ledger)),
        ):
            assert await resolver.resolve(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_resolve_x_no_ledger(self, profile: Profile, resolver: IndyDIDResolver):
        """Test resolve method with no ledger."""
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, None)
        )
        profile.context.injector.bind_instance(IndyLedgerRequestsExecutor, mock_executor)
        with pytest.raises(ResolverError):
            await resolver.resolve(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_resolve_x_did_not_found(
        self, resolver: IndyDIDResolver, ledger: BaseLedger, profile: Profile
    ):
        """Test resolve method when no did is found."""
        ledger.get_key_for_did.side_effect = LedgerError
        with pytest.raises(DIDNotFound):
            await resolver.resolve(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_supports_updated_did_sov_rules(
        self, resolver: IndyDIDResolver, ledger: BaseLedger, profile: Profile
    ):
        """Test that new attrib structure is supported."""
        example = {
            "endpoint": "https://example.com/endpoint",
            "routingKeys": ["HQhjaj4mcaS3Xci27a9QhnBrNpS91VNFUU4TDrtMxa9j"],
            "types": ["DIDComm", "did-communication", "endpoint"],
            "profile": "https://example.com",
            "linked_domains": "https://example.com",
        }

        ledger.get_all_endpoints_for_did = mock.CoroutineMock(return_value=example)
        assert await resolver.resolve(profile, TEST_DID0)

    @pytest.mark.asyncio
    async def test_supports_updated_did_sov_rules_no_endpoint_url(
        self, resolver: IndyDIDResolver, ledger: BaseLedger, profile: Profile
    ):
        """Test that new attrib structure is supported."""
        example = {
            "routingKeys": ["a-routing-key"],
            "types": ["DIDComm", "did-communication", "endpoint"],
        }

        ledger.get_all_endpoints_for_did = mock.CoroutineMock(return_value=example)
        result = await resolver.resolve(profile, TEST_DID0)
        assert "service" not in result

    @pytest.mark.parametrize(
        "types, result",
        [
            (
                [],
                ["endpoint", "did-communication"],
            ),
            (
                ["did-communication"],
                ["did-communication"],
            ),
            (
                ["endpoint", "did-communication", "DIDComm", "other-endpoint-type"],
                ["endpoint", "did-communication"],
            ),
            (
                ["endpoint", "did-communication", "DIDComm"],
                ["endpoint", "did-communication", "DIDComm"],
            ),
        ],
    )
    def test_process_endpoint_types(self, resolver: IndyDIDResolver, types, result):
        assert resolver.process_endpoint_types(types) == result

    @pytest.mark.parametrize(
        "keys",
        [
            ["3YJCx3TqotDWFGv7JMR5erEvrmgu5y4FDqjR7sKWxgXn"],
            ["did:key:z6MkgzZFYHiH9RhyMmkoyvNvVwnvgLxkVrJbureLx9HXsuKA"],
            [
                "did:key:z6MkgzZFYHiH9RhyMmkoyvNvVwnvgLxkVrJbureLx9HXsuKA#z6MkgzZFYHiH9RhyMmkoyvNvVwnvgLxkVrJbureLx9HXsuKA"
            ],
        ],
    )
    def test_routing_keys_as_did_key_urls(self, keys):
        for key in _routing_keys_as_did_key_urls(keys):
            assert key.startswith("did:key:")
            assert "#" in key



"""Wallet configuration."""

import logging
from typing import Tuple

from ..core.error import ProfileNotFoundError, StartupError
from ..core.profile import Profile, ProfileManager, ProfileSession
from ..storage.base import BaseStorage
from ..storage.error import StorageNotFoundError
from ..version import RECORD_TYPE_ACAPY_VERSION, __version__
from ..wallet.base import BaseWallet
from ..wallet.crypto import seed_to_did
from ..wallet.did_info import DIDInfo
from ..wallet.did_method import SOV
from ..wallet.key_type import ED25519
from .base import ConfigError
from .injection_context import InjectionContext

LOGGER = logging.getLogger(__name__)

CFG_MAP = {
    "key",
    "key_derivation_method",
    "rekey",
    "name",
    "storage_config",
    "storage_creds",
    "storage_type",
    "test",
}


def _create_config_with_settings(settings) -> dict:
    profile_config = {}

    for k in CFG_MAP:
        pk = f"wallet.{k}"
        if pk in settings:
            profile_config[k] = settings[pk]

    # may be set by `aca-py provision --recreate`
    if settings.get("wallet.recreate"):
        profile_config["auto_recreate"] = True

    return profile_config


async def _attempt_open_profile(
    profile_manager: ProfileManager,
    context: InjectionContext,
    profile_config: dict,
    settings: dict,
) -> Tuple[Profile, bool]:
    provision = False
    try:
        profile = await profile_manager.open(context, profile_config)
    except ProfileNotFoundError:
        if settings.get("auto_provision", False):
            profile = await profile_manager.provision(context, profile_config)
            provision = True
        else:
            error_msg = (
                "Profile not found. Use `aca-py start --auto-provision` to create."
            )
            LOGGER.error(error_msg)
            raise StartupError(error_msg)

    return (profile, provision)


async def _replace_public_did_if_seed_mismatch(
    public_did_info: DIDInfo,
    wallet: BaseWallet,
    settings: dict,
    wallet_seed: str,
) -> DIDInfo:
    """Replace the public DID if a seed is provided and doesn't match the current DID.

    Args:
        public_did_info: Current public DID info
        wallet: Wallet instance
        settings: Configuration settings
        wallet_seed: Optional seed to check against current DID

    Returns:
        DIDInfo: Either the original DID info or a new one if replaced
    """
    if not wallet_seed:
        return public_did_info

    public_did = public_did_info.did
    if seed_to_did(wallet_seed) == public_did:
        return public_did_info

    if not settings.get("wallet.replace_public_did"):
        raise ConfigError(
            "New seed provided which doesn't match the registered "
            f"public did {public_did}"
        )

    LOGGER.info(
        "Replacing public DID which doesn't match the seed "
        "(as configured by --replace-public-did)"
    )
    replace_did_info = await wallet.create_local_did(
        method=SOV, key_type=ED25519, seed=wallet_seed
    )
    await wallet.set_public_did(replace_did_info.did)
    LOGGER.info(
        "Created new public DID: %s, with verkey: %s",
        replace_did_info.did,
        replace_did_info.verkey,
    )
    return replace_did_info


async def _initialize_with_debug_settings(settings: dict, wallet: BaseWallet):
    test_seed = settings.get("debug.seed")
    if settings.get("debug.enabled") and not test_seed:
        test_seed = "testseed000000000000000000000001"
    if test_seed:
        await wallet.create_local_did(
            method=SOV,
            key_type=ED25519,
            seed=test_seed,
            metadata={"endpoint": "1.2.3.4:8021"},
        )


async def _initialize_with_seed(
    settings: dict, wallet: BaseWallet, create_local_did: bool, seed: str
) -> DIDInfo:
    if create_local_did:
        endpoint = settings.get("default_endpoint")
        metadata = {"endpoint": endpoint} if endpoint else None

        did_info = await wallet.create_local_did(
            method=SOV,
            key_type=ED25519,
            seed=seed,
            metadata=metadata,
        )
    else:
        did_info = await wallet.create_public_did(method=SOV, key_type=ED25519, seed=seed)

    LOGGER.info(
        "Created new %s DID: %s, Verkey: %s",
        "local" if create_local_did else "public",
        did_info.did,
        did_info.verkey,
    )
    return did_info


async def wallet_config(
    context: InjectionContext, provision: bool = False
) -> Tuple[Profile, DIDInfo]:
    """Initialize the root profile."""

    profile_manager = context.inject(ProfileManager)

    settings = context.settings
    profile_config = _create_config_with_settings(settings)
    wallet_seed = settings.get("wallet.seed")
    create_local_did = settings.get("wallet.local_did")

    if provision:
        profile = await profile_manager.provision(context, profile_config)
    else:
        profile, provision = await _attempt_open_profile(
            profile_manager, context, profile_config, settings
        )

    LOGGER.info(
        "%s Profile name: %s, backend: %s",
        "Created new profile -" if profile.created else "Opened existing profile -",
        profile.name,
        profile.backend,
    )

    txn = await profile.transaction()
    wallet = txn.inject(BaseWallet)
    public_did_info = await wallet.get_public_did()

    if public_did_info:
        # Check if we need to replace the public DID due to seed mismatch
        public_did_info = await _replace_public_did_if_seed_mismatch(
            public_did_info, wallet, settings, wallet_seed
        )
    elif wallet_seed:
        # Create new public DID from seed if none exists
        public_did_info = await _initialize_with_seed(
            settings, wallet, create_local_did, wallet_seed
        )

    public_did = public_did_info.did if public_did_info else None
    if provision and not create_local_did and not public_did:
        LOGGER.info("No public DID created")

    await _initialize_with_debug_settings(settings, wallet)

    await txn.commit()

    return (profile, public_did_info)


async def add_or_update_version_to_storage(session: ProfileSession):
    """Add or update ACA-Py version StorageRecord."""
    storage: BaseStorage = session.inject(BaseStorage)
    try:
        record = await storage.find_record(RECORD_TYPE_ACAPY_VERSION, {})
        await storage.update_record(record, f"v{__version__}", {})
    except StorageNotFoundError:
        raise ConfigError(
            (
                "No wallet storage version found, Run aca-py "
                "upgrade command with --from-version argument "
                "to fix this."
            )
        )



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import PROTOCOL_PACKAGE, QUERIES
from ..queries import Queries, QueryItem

TEST_QUERY_PROTOCOL = "https://didcomm.org/tictactoe/1.*"
TEST_QUERY_GOAL_CODE = "aries.*"


class TestQueries(TestCase):
    def test_init(self):
        test_queries = [
            QueryItem(feature_type="protocol", match=TEST_QUERY_PROTOCOL),
            QueryItem(feature_type="goal-code", match=TEST_QUERY_GOAL_CODE),
        ]
        queries = Queries(queries=test_queries)
        assert queries.queries[0].feature_type == "protocol"
        assert queries.queries[0].match == TEST_QUERY_PROTOCOL
        assert queries.queries[1].feature_type == "goal-code"
        assert queries.queries[1].match == TEST_QUERY_GOAL_CODE
        test_queries = [
            QueryItem(feature_type="protocol", match=TEST_QUERY_PROTOCOL),
        ]
        queries = Queries(queries=test_queries)
        assert queries.queries[0].feature_type == "protocol"
        assert queries.queries[0].match == TEST_QUERY_PROTOCOL

    def test_type(self):
        test_queries = [
            QueryItem(feature_type="protocol", match=TEST_QUERY_PROTOCOL),
            QueryItem(feature_type="goal-code", match=TEST_QUERY_GOAL_CODE),
        ]
        queries = Queries(queries=test_queries)
        assert queries._type == DIDCommPrefix.qualify_current(QUERIES)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.queries.QueriesSchema.load")
    def test_deserialize(self, mock_queries_schema_dump):
        obj = {"obj": "obj"}

        query = Queries.deserialize(obj)
        mock_queries_schema_dump.assert_called_once_with(obj)

        assert query is mock_queries_schema_dump.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.queries.QueriesSchema.dump")
    def test_serialize(self, mock_queries_schema_dump):
        test_queries = [
            QueryItem(feature_type="protocol", match=TEST_QUERY_PROTOCOL),
            QueryItem(feature_type="goal-code", match=TEST_QUERY_GOAL_CODE),
        ]
        queries = Queries(queries=test_queries)

        queries_dict = queries.serialize()
        mock_queries_schema_dump.assert_called_once_with(queries)

        assert queries_dict is mock_queries_schema_dump.return_value


class TestQuerySchema(TestCase):
    test_queries = [
        QueryItem(feature_type="protocol", match=TEST_QUERY_PROTOCOL),
        QueryItem(feature_type="goal-code", match=TEST_QUERY_GOAL_CODE),
    ]
    queries = Queries(queries=test_queries)

    def test_make_model(self):
        data = self.queries.serialize()
        model_instance = Queries.deserialize(data)
        assert isinstance(model_instance, Queries)



from unittest import TestCase

from ...models.cred_detail import LDProofVCDetail
from ...models.cred_detail_options import LDProofVCDetailOptions

VC_DETAIL_OPTIONS = {
    "proofType": "Ed25519Signature2018",
    "proofPurpose": "assertionMethod",
    "created": "2010-01-01T19:73:24Z",
    "domain": "example.com",
    "challenge": "019a5321-1c1f-4031-8b9b-8c0234be7d78",
    "credentialStatus": {"type": "CredentialStatusList2017"},
}

VC_DETAIL_CREDENTIAL = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/citizenship/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
    "type": ["VerifiableCredential", "PermanentResidentCard"],
    "issuer": "did:example:489398593",
    "identifier": "83627465",
    "name": "Permanent Resident Card",
    "description": "Government of Example Permanent Resident Card.",
    "issuanceDate": "2019-12-03T12:19:52Z",
    "expirationDate": "2029-12-03T12:19:52Z",
    "credentialSubject": {
        "id": "did:example:b34ca6cd37bbf23",
        "type": ["PermanentResident", "Person"],
        "givenName": "JOHN",
        "familyName": "SMITH",
        "gender": "Male",
        "image": "data:image/png;base64,iVBORw0KGgokJggg==",
        "residentSince": "2015-01-01",
        "lprCategory": "C09",
        "lprNumber": "999-999-999",
        "commuterClassification": "C1",
        "birthCountry": "Bahamas",
        "birthDate": "1958-07-17",
    },
}

VC_DETAIL = {"credential": VC_DETAIL_CREDENTIAL, "options": VC_DETAIL_OPTIONS}


class TestLDProofVCDetail(TestCase):
    """LDProofVCDetail correctness proof tests"""

    def test_serde(self):
        """Test de/serialization."""
        detail = LDProofVCDetail.deserialize(VC_DETAIL)
        assert type(detail) is LDProofVCDetail

        detail_dict = detail.serialize()
        assert detail_dict == VC_DETAIL


class TestLDProofVCDetailOptions(TestCase):
    """LDProofVCDetail correctness proof tests"""

    def test_serde(self):
        """Test de/serialization."""
        detail_options = LDProofVCDetailOptions.deserialize(VC_DETAIL_OPTIONS)
        assert type(detail_options) is LDProofVCDetailOptions

        detail_options_dict = detail_options.serialize()
        assert detail_options_dict == VC_DETAIL_OPTIONS



"""mediate-grant message.

Used to notify mediation client of a granted mediation request.
"""

from typing import Optional, Sequence

from marshmallow import fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import MEDIATE_GRANT, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.mediation_grant_handler.MediationGrantHandler"
)


class MediationGrant(AgentMessage):
    """Class representing a mediation grant message."""

    class Meta:
        """Metadata for a mediation grant."""

        handler_class = HANDLER_CLASS
        message_type = MEDIATE_GRANT
        schema_class = "MediationGrantSchema"

    def __init__(
        self,
        *,
        endpoint: Optional[str] = None,
        routing_keys: Optional[Sequence[str]] = None,
        **kwargs,
    ):
        """Initialize mediation grant object.

        Args:
            endpoint: Endpoint address for the mediation route
            routing_keys: Keys for the mediation route
            kwargs: Additional keyword arguments for the message
        """
        super(MediationGrant, self).__init__(**kwargs)
        self.endpoint = endpoint
        self.routing_keys = routing_keys or []


class MediationGrantSchema(AgentMessageSchema):
    """Mediation grant schema class."""

    class Meta:
        """Mediation grant schema metadata."""

        model_class = MediationGrant

    endpoint = fields.Str(
        metadata={
            "description": (
                "endpoint on which messages destined for the recipient are received."
            ),
            "example": "http://192.168.56.102:8020/",
        }
    )
    routing_keys = fields.List(
        fields.Str(metadata={"description": "Keys to use for forward message packaging"}),
        required=False,
    )



"""Injection context implementation."""

import copy
from collections import namedtuple
from typing import Mapping, Optional, Type

from .base import BaseInjector, InjectionError
from .injector import Injector, InjectType
from .settings import Settings

Scope = namedtuple("Scope", "name injector")


class InjectionContextError(InjectionError):
    """Base class for issues in the injection context."""


class InjectionContext(BaseInjector):
    """Manager for configuration settings and class providers."""

    ROOT_SCOPE = "application"

    def __init__(
        self,
        *,
        settings: Optional[Mapping[str, object]] = None,
        enforce_typing: bool = True,
    ):
        """Initialize a `ServiceConfig`."""
        self._injector = Injector(settings, enforce_typing=enforce_typing)
        self._scope_name = InjectionContext.ROOT_SCOPE

    @property
    def injector(self) -> Injector:
        """Accessor for scope-specific injector."""
        return self._injector

    @injector.setter
    def injector(self, injector: Injector):
        """Setter for scope-specific injector."""
        self._injector = injector

    @property
    def settings(self) -> Settings:
        """Accessor for scope-specific settings."""
        return self.injector.settings

    @settings.setter
    def settings(self, settings: Settings):
        """Setter for scope-specific settings."""
        self.injector.settings = settings

    def update_settings(self, settings: Mapping[str, object]):
        """Update the scope with additional settings."""
        if settings:
            self.injector.settings.update(settings)

    def start_scope(
        self, settings: Optional[Mapping[str, object]] = None
    ) -> "InjectionContext":
        """Begin a new named scope.

        Args:
            scope_name: The unique name for the scope being entered
            settings: An optional mapping of additional settings to apply

        Returns:
            A new injection context representing the scope

        """
        result = self.copy()
        if settings:
            result.update_settings(settings)
        return result

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, object]] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            base_cls (Type[InjectType]): The base class to retrieve an instance of.
            settings (Optional[Mapping[str, object]]): An optional mapping providing
                configuration to the provider.

        Returns:
            InjectType: An instance of the base class, or None.

        """
        return self.injector.inject(base_cls, settings)

    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, object]] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """
        return self.injector.inject_or(base_cls, settings, default)

    def copy(self) -> "InjectionContext":
        """Produce a copy of the injector instance."""
        result = copy.copy(self)
        result._injector = self.injector.copy()
        return result



"""Utilities for Processing Replies to Domain Read Requests."""

import base64
import hashlib
import json
from binascii import hexlify
from copy import deepcopy

import base58

from .constants import (
    ACCUM_FROM,
    ACCUM_TO,
    ALL_ATR_KEYS,
    ATTR_NAMES,
    ATTRIB,
    AUDIT_PATH,
    CLAIM_DEF,
    CRED_DEF_ID,
    DATA,
    DEST,
    FROM,
    GET_ATTR,
    GET_CLAIM_DEF,
    GET_NYM,
    GET_REVOC_REG_DEF,
    GET_REVOC_REG_DELTA,
    GET_REVOC_REG_ENTRY,
    GET_SCHEMA,
    HASH,
    LAST_SEQ_NO,
    LAST_UPDATE_TIME,
    MARKER_ATTR,
    MARKER_CLAIM_DEF,
    MARKER_REVOC_DEF,
    MARKER_REVOC_REG_ENTRY,
    MARKER_REVOC_REG_ENTRY_ACCUM,
    MARKER_SCHEMA,
    METADATA,
    NAME,
    NYM,
    PROOF_NODES,
    REF,
    RESULT,
    REVOC_DEF_TYPE,
    REVOC_DEF_TYPE_ID,
    REVOC_REG_DEF,
    REVOC_REG_ENTRY,
    ROOT_HASH,
    SCHEMA,
    SEQ_NO,
    STATE_PROOF,
    STATE_PROOF_FROM,
    TAG,
    TXN,
    TXN_METADATA,
    TXN_TIME,
    VAL,
    VALUE,
    VERSION,
)
from .utils import audit_path_length


def _extract_attr_typed_value(txn_data):
    """Check for 'raw', 'enc', 'hash' in ATTR & GET_ATTR, return it's name and value."""
    existing_keys = [key for key in ALL_ATR_KEYS if key in txn_data]
    if len(existing_keys) == 0:
        raise ValueError(
            "ATTR should have one of the following fields: {}".format(ALL_ATR_KEYS)
        )
    if len(existing_keys) > 1:
        raise ValueError(
            "ATTR should have only one of the following fields: {}".format(ALL_ATR_KEYS)
        )
    existing_key = existing_keys[0]
    return existing_key, txn_data[existing_key]


def parse_attr_txn(txn_data):
    """Process txn_data and parse attr_txn based on attr_type."""
    attr_type, attr = _extract_attr_typed_value(txn_data)
    if attr_type == "raw":
        data = json.loads(attr)
        re_raw = json.dumps(data)
        key, _ = data.popitem()
        return attr_type, key, re_raw
    if attr_type == "enc":
        return attr_type, attr, attr
    if attr_type == "hash":
        return attr_type, attr, None


def encode_state_value(value, seqNo, txnTime):
    """Return encoded state value."""
    return json.dumps({LAST_SEQ_NO: seqNo, LAST_UPDATE_TIME: txnTime, VAL: value})


def decode_state_value(encoded_value):
    """Return val, lsn, lut from encoded state value."""
    decoded = json.loads(encoded_value)
    value = decoded.get(VAL)
    last_seq_no = decoded.get(LAST_SEQ_NO)
    last_update_time = decoded.get(LAST_UPDATE_TIME)
    return value, last_seq_no, last_update_time


def hash_of(text) -> str:
    """Return 256 bit hexadecimal digest of text."""
    if not isinstance(text, (str, bytes)):
        text = json.dumps(text)
    if not isinstance(text, bytes):
        text = text.encode()
    return hashlib.sha256(text).hexdigest()


def make_state_path_for_attr(did, attr_name, attr_is_hash=False) -> bytes:
    """Return state_path for ATTR."""
    nameHash = (
        hashlib.sha256(attr_name.encode()).hexdigest() if not attr_is_hash else attr_name
    )
    return "{DID}:{MARKER}:{ATTR_NAME}".format(
        DID=did, MARKER=MARKER_ATTR, ATTR_NAME=nameHash
    ).encode()


def prepare_get_attr_for_state(reply):
    """Return value for state from GET_ATTR."""
    attr_type, attr_key = _extract_attr_typed_value(reply)
    data = reply.get(DATA)
    value_bytes = None
    if data:
        result = reply.copy()
        data = result.pop(DATA)
        result[attr_type] = data
        attr_type, _, value = parse_attr_txn(result)
        hashed_value = hash_of(value) if value else ""
        seq_no = result.get(SEQ_NO)
        txn_time = result.get(TXN_TIME)
        value_bytes = encode_state_value(hashed_value, seq_no, txn_time)
    return value_bytes


def prepare_attr_for_state(txn, path_only=False):
    """Return key, value pair for state from ATTR."""
    txn_data = txn.get(TXN, {}).get(DATA)
    nym = txn_data.get(DEST)
    attr_type, attr_key, value = parse_attr_txn(txn_data)
    path = make_state_path_for_attr(nym, attr_key, attr_type == HASH)
    if path_only:
        return path
    hashed_value = hash_of(value) if value else ""
    seq_no = txn[TXN_METADATA].get(SEQ_NO)
    txn_time = txn[TXN_METADATA].get(TXN_TIME)
    value_bytes = encode_state_value(hashed_value, seq_no, txn_time)
    return path, value_bytes.encode()


def make_state_path_for_nym(did) -> bytes:
    """Return state_path for NYM."""
    return hashlib.sha256(did.encode()).digest()


def prepare_nym_for_state(txn):
    """Return encoded state path from NYM."""
    txn_data = txn.get(TXN, {}).get(DATA)
    nym = txn_data.get(DEST)
    path = make_state_path_for_nym(nym)
    return hexlify(path).decode()


def prepare_get_nym_for_state(reply):
    """Return value for state from GET_NYM."""
    data = reply.get(DATA)
    value = None
    if data is not None:
        if isinstance(data, str):
            data = json.loads(data)
        data.pop(DEST, None)
        value = json.dumps(data)
    return value


def prepare_get_schema_for_state(reply):
    """Return value for state from GET_SCHEMA."""
    value_bytes = None
    attr_names = reply[DATA].get(ATTR_NAMES)
    if attr_names:
        data = {ATTR_NAMES: attr_names}
        seq_no = reply.get(SEQ_NO)
        txn_time = reply.get(TXN_TIME)
        value_bytes = encode_state_value(data, seq_no, txn_time)
    return value_bytes


def make_state_path_for_schema(authors_did, schema_name, schema_version) -> bytes:
    """Return state_path for SCHEMA."""
    return "{DID}:{MARKER}:{SCHEMA_NAME}:{SCHEMA_VERSION}".format(
        DID=authors_did,
        MARKER=MARKER_SCHEMA,
        SCHEMA_NAME=schema_name,
        SCHEMA_VERSION=schema_version,
    ).encode()


def prepare_schema_for_state(txn, path_only=False):
    """Return key-value pair for state from SCHEMA."""
    origin = txn[TXN].get(METADATA, {}).get(FROM)
    schema_name = txn[TXN][DATA][DATA].get(NAME)
    schema_version = txn[TXN][DATA][DATA].get(VERSION)
    value = {ATTR_NAMES: txn[TXN][DATA][DATA].get(ATTR_NAMES)}
    path = make_state_path_for_schema(origin, schema_name, schema_version)
    if path_only:
        return path
    seq_no = txn[TXN_METADATA].get(SEQ_NO)
    txn_time = txn[TXN_METADATA].get(TXN_TIME)
    value_bytes = encode_state_value(value, seq_no, txn_time)
    return path, value_bytes.encode()


def prepare_get_claim_def_for_state(reply):
    """Return value for state from GET_CLAIM_DEF."""
    schema_seq_no = reply.get(REF)
    if schema_seq_no is None:
        raise ValueError("ref field is absent, but it must contain schema seq no")
    value_bytes = None
    data = reply.get(DATA)
    if data is not None:
        seq_no = reply.get(SEQ_NO)
        txn_time = reply.get(TXN_TIME)
        value_bytes = encode_state_value(data, seq_no, txn_time)
    return value_bytes


def make_state_path_for_claim_def(authors_did, schema_seq_no, signature_type, tag):
    """Return state_path for CLAIM DEF."""
    return "{DID}:{MARKER}:{SIGNATURE_TYPE}:{SCHEMA_SEQ_NO}:{TAG}".format(
        DID=authors_did,
        MARKER=MARKER_CLAIM_DEF,
        SIGNATURE_TYPE=signature_type,
        SCHEMA_SEQ_NO=schema_seq_no,
        TAG=tag,
    ).encode()


def prepare_claim_def_for_state(txn, path_only=False):
    """Return key-value pair for state from CLAIM_DEF."""
    origin = txn[TXN].get(METADATA, {}).get(FROM)
    schema_seq_no = txn[TXN][DATA].get(REF)
    if schema_seq_no is None:
        raise ValueError("ref field is absent, but it must contain schema seq no")
    data = txn[TXN][DATA].get(DATA)
    if data is None:
        raise ValueError("data field is absent, but it must contain components of keys")
    signature_type = txn[TXN][DATA].get("signature_type", "CL")
    tag = txn[TXN][DATA].get(TAG, "tag")
    path = make_state_path_for_claim_def(origin, schema_seq_no, signature_type, tag)
    if path_only:
        return path
    seq_no = txn[TXN_METADATA].get(SEQ_NO)
    txn_time = txn[TXN_METADATA].get(TXN_TIME)
    value_bytes = encode_state_value(data, seq_no, txn_time)
    return path, value_bytes.encode()


def prepare_get_revoc_def_for_state(reply):
    """Return value for state from GET_REVOC_DEF."""
    seq_no = reply.get(SEQ_NO)
    txn_time = reply.get(TXN_TIME)
    value_bytes = encode_state_value(reply.get(DATA), seq_no, txn_time)
    return value_bytes


def make_state_path_for_revoc_def(
    authors_did, cred_def_id, revoc_def_type, revoc_def_tag
) -> bytes:
    """Return state_path for REVOC_DEF."""
    return "{DID}:{MARKER}:{CRED_DEF_ID}:{REVOC_DEF_TYPE}:{REVOC_DEF_TAG}".format(
        DID=authors_did,
        MARKER=MARKER_REVOC_DEF,
        CRED_DEF_ID=cred_def_id,
        REVOC_DEF_TYPE=revoc_def_type,
        REVOC_DEF_TAG=revoc_def_tag,
    ).encode()


def prepare_revoc_def_for_state(txn, path_only=False):
    """Return key-value pair for state from REVOC_DEF."""
    author_did = txn[TXN].get(METADATA, {}).get(FROM, None)
    txn_data = txn[TXN].get(DATA)
    cred_def_id = txn_data.get(CRED_DEF_ID)
    revoc_def_type = txn_data.get(REVOC_DEF_TYPE)
    revoc_def_tag = txn_data.get(TAG)
    path = make_state_path_for_revoc_def(
        author_did, cred_def_id, revoc_def_type, revoc_def_tag
    )
    if path_only:
        return path
    seq_no = txn[TXN_METADATA].get(SEQ_NO)
    txn_time = txn[TXN_METADATA].get(TXN_TIME)
    value_bytes = encode_state_value(txn_data, seq_no, txn_time)
    return path, value_bytes.encode()


def prepare_get_revoc_reg_entry_for_state(reply):
    """Return value for state from GET_REVOC_REG_ENTRY."""
    if RESULT in reply.keys():
        result = reply.get(RESULT)
    else:
        result = reply
    seq_no = result.get(SEQ_NO)
    txn_time = result.get(TXN_TIME)
    value_bytes = encode_state_value(result.get(DATA), seq_no, txn_time)
    return value_bytes


def make_state_path_for_revoc_reg_entry(revoc_reg_def_id) -> bytes:
    """Return state_path for REVOC_REG_ENTRY."""
    return "{MARKER}:{REVOC_REG_DEF_ID}".format(
        MARKER=MARKER_REVOC_REG_ENTRY, REVOC_REG_DEF_ID=revoc_reg_def_id
    ).encode()


def prepare_get_revoc_reg_delta_for_state(reply):
    """Return value for state from GET_REVOC_REG_DELTA."""
    if STATE_PROOF_FROM in reply[DATA]:
        accum_to_seq_no = reply[DATA][VALUE].get(ACCUM_TO, {}).get(SEQ_NO)
        accum_to_txn_time = reply[DATA][VALUE].get(ACCUM_TO, {}).get(TXN_TIME)
        accum_from_seq_no = reply[DATA][VALUE].get(ACCUM_FROM, {}).get(SEQ_NO)
        accum_from_txn_time = reply[DATA][VALUE].get(ACCUM_FROM, {}).get(TXN_TIME)
        return (
            encode_state_value(
                reply[DATA][VALUE].get(ACCUM_TO),
                accum_to_seq_no,
                accum_to_txn_time,
            ),
            encode_state_value(
                reply[DATA][VALUE].get(ACCUM_FROM),
                accum_from_seq_no,
                accum_from_txn_time,
            ),
        )
    else:
        seq_no = reply.get(SEQ_NO)
        txn_time = reply.get(TXN_TIME)
        return encode_state_value(reply[DATA][VALUE].get(ACCUM_TO), seq_no, txn_time)


def prepare_revoc_reg_entry_for_state(txn, path_only=False):
    """Return key-value pair for state from REVOC_REG_ENTRY."""
    txn_data = txn[TXN].get(DATA)
    revoc_reg_def_id = txn_data.get(REVOC_DEF_TYPE_ID)
    path = make_state_path_for_revoc_reg_entry(revoc_reg_def_id=revoc_reg_def_id)
    if path_only:
        return path
    seq_no = txn[TXN_METADATA].get(SEQ_NO)
    txn_time = txn[TXN_METADATA].get(TXN_TIME)
    txn_data = deepcopy(txn_data)
    txn_data[SEQ_NO] = seq_no
    txn_data[TXN_TIME] = txn_time
    value_bytes = encode_state_value(txn_data, seq_no, txn_time)
    return path, value_bytes.encode()


def prepare_get_revoc_reg_entry_accum_for_state(reply):
    """Return value for state from GET_REVOC_REG_ENTRY_ACCUM."""
    seq_no = reply.get(SEQ_NO)
    txn_time = reply.get(TXN_TIME)
    value_bytes = encode_state_value(reply.get(DATA), seq_no, txn_time)
    return value_bytes


def make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id) -> bytes:
    """Return state_path for REVOC_REG_ENTRY_ACCUM."""
    return "{MARKER}:{REVOC_REG_DEF_ID}".format(
        MARKER=MARKER_REVOC_REG_ENTRY_ACCUM, REVOC_REG_DEF_ID=revoc_reg_def_id
    ).encode()


def prepare_revoc_reg_entry_accum_for_state(txn):
    """Return key-value pair for state from REVOC_REG_ENTRY_ACCUM."""
    if RESULT in txn.keys():
        result = txn.get(RESULT)
    else:
        result = txn
    txn_data = result[TXN].get(DATA)
    revoc_reg_def_id = txn_data.get(REVOC_DEF_TYPE_ID)
    seq_no = result[TXN_METADATA].get(SEQ_NO)
    txn_time = result[TXN_METADATA].get(TXN_TIME)
    path = make_state_path_for_revoc_reg_entry_accum(revoc_reg_def_id=revoc_reg_def_id)
    txn_data = deepcopy(txn_data)
    txn_data[SEQ_NO] = seq_no
    txn_data[TXN_TIME] = txn_time
    value_bytes = encode_state_value(txn_data, seq_no, txn_time)
    return path, value_bytes.encode()


def extract_params_write_request(data):
    """Return tree_size, leaf_index, audit_path, expected_root_hash from reply."""
    if RESULT in data.keys():
        data = data.get(RESULT)
    tree_size = data[TXN_METADATA].get(SEQ_NO)
    leaf_index = tree_size - 1
    audit_path = data.get(AUDIT_PATH)
    audit_path = audit_path[:]
    decoded_audit_path = [
        base58.b58decode(hash_str.encode("utf-8")) for hash_str in audit_path
    ]
    expected_root_hash = base58.b58decode(data.get(ROOT_HASH).encode("utf-8"))
    if len(decoded_audit_path) != audit_path_length(leaf_index, tree_size):
        raise Exception("auditPath length does not match with given seqNo")
    return tree_size, leaf_index, decoded_audit_path, expected_root_hash


def get_proof_nodes(reply):
    """Return proof_nodes from reply."""
    if RESULT in reply.keys():
        reply = reply.get(RESULT)
    if reply.get("type") == GET_REVOC_REG_DELTA and STATE_PROOF_FROM in reply.get(DATA):
        proof_nodes_accum_to = reply[STATE_PROOF].get(PROOF_NODES)
        proof_nodes_accum_from = reply[DATA].get(STATE_PROOF_FROM, {}).get(PROOF_NODES)
        return base64.b64decode(proof_nodes_accum_to), base64.b64decode(
            proof_nodes_accum_from
        )
    else:
        b64_encoded_nodes = reply[STATE_PROOF].get(PROOF_NODES)
        return base64.b64decode(b64_encoded_nodes)


def prepare_for_state_read(reply):
    """Return state value from read requests reply."""
    if RESULT in reply.keys():
        reply = reply.get(RESULT)
    request_type = reply.get("type")
    if request_type == GET_ATTR:
        return prepare_get_attr_for_state(reply=reply)
    if request_type == GET_NYM:
        return prepare_get_nym_for_state(reply=reply)
    if request_type == GET_SCHEMA:
        return prepare_get_schema_for_state(reply=reply)
    if request_type == GET_CLAIM_DEF:
        return prepare_get_claim_def_for_state(reply=reply)
    if request_type == GET_REVOC_REG_DEF:
        return prepare_get_revoc_def_for_state(reply=reply)
    if request_type == GET_REVOC_REG_ENTRY:
        return prepare_get_revoc_reg_entry_accum_for_state(reply=reply)
    if request_type == GET_REVOC_REG_DELTA:
        if "issued" in reply[DATA].get("value"):
            return prepare_get_revoc_reg_delta_for_state(reply=reply)
        else:
            return prepare_get_revoc_reg_entry_accum_for_state(reply=reply)
    raise ValueError(
        "Cannot make state value for request of type {}".format(request_type)
    )


def prepare_for_state_write(reply):
    """Return state key, value pair from write requests reply."""
    if RESULT in reply.keys():
        reply = reply.get(RESULT)
    request_type = reply[TXN].get("type")
    if request_type == NYM:
        return prepare_nym_for_state(txn=reply)
    if request_type == ATTRIB:
        return prepare_attr_for_state(txn=reply)
    if request_type == SCHEMA:
        return prepare_schema_for_state(txn=reply)
    if request_type == CLAIM_DEF:
        return prepare_claim_def_for_state(txn=reply)
    if request_type == REVOC_REG_DEF:
        return prepare_revoc_def_for_state(txn=reply)
    if request_type == REVOC_REG_ENTRY:
        return prepare_revoc_reg_entry_for_state(txn=reply)
    raise ValueError(
        "Cannot make state key-value pair for request of type {}".format(request_type)
    )



from unittest import mock

import pytest

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import PROBLEM_REPORT
from .. import problem_report as test_module
from ..problem_report import DIDXProblemReport

THID = "dummy-thid"
PTHID = "dummy-pthid"


def test_init_type():
    complete = DIDXProblemReport()
    assert complete._type == DIDCommPrefix.qualify_current(PROBLEM_REPORT)


def test_serde():
    obj = {
        "~thread": {"thid": THID, "pthid": PTHID},
        "description": {"code": "complete_not_accepted", "en": "test"},
    }
    report = DIDXProblemReport.deserialize(obj)
    assert report._type == DIDCommPrefix.qualify_current(PROBLEM_REPORT)
    complete_dict = report.serialize()
    assert complete_dict["~thread"] == obj["~thread"]


def test_missing_code():
    with pytest.raises(BaseModelError):
        DIDXProblemReport.deserialize({"description": {"en": "test"}})


def test_unrecognized_code():
    with mock.patch.object(test_module, "LOGGER", autospec=True) as mock_logger:
        DIDXProblemReport.deserialize({"description": {"code": "unknown", "en": "test"}})
    mock_logger.warning.assert_called_once()



from unittest import IsolatedAsyncioTestCase

from ..error import BaseError


class TestBaseError(IsolatedAsyncioTestCase):
    async def test_base_error(self):
        err = BaseError()
        assert err.error_code is None
        assert not err.message
        assert err.roll_up == f"{err.__class__.__name__}."

        MESSAGE = "Not enough space\nClear 10MB\n\n"
        CODE = "-1"
        err = BaseError(MESSAGE, error_code=CODE)
        assert err.error_code == CODE
        assert err.message == MESSAGE.strip()
        assert err.roll_up == "Not enough space. Clear 10MB."

        zdx = ZeroDivisionError()
        keyx = KeyError("world")
        osx = OSError("oh\nno")
        iox = IOError("hello")
        keyx.__cause__ = zdx
        osx.__cause__ = keyx
        iox.__cause__ = osx
        err.__cause__ = iox

        assert err.roll_up == (
            "Not enough space. Clear 10MB. hello. oh. no. world. ZeroDivisionError."
        )



ED25519_2020_V1 = {
    "@context": {
        "id": "@id",
        "type": "@type",
        "@protected": True,
        "proof": {
            "@id": "https://w3id.org/security#proof",
            "@type": "@id",
            "@container": "@graph",
        },
        "Ed25519VerificationKey2020": {
            "@id": "https://w3id.org/security#Ed25519VerificationKey2020",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "controller": {
                    "@id": "https://w3id.org/security#controller",
                    "@type": "@id",
                },
                "revoked": {
                    "@id": "https://w3id.org/security#revoked",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "publicKeyMultibase": {
                    "@id": "https://w3id.org/security#publicKeyMultibase",
                    "@type": "https://w3id.org/security#multibase",
                },
            },
        },
        "Ed25519Signature2020": {
            "@id": "https://w3id.org/security#Ed25519Signature2020",
            "@context": {
                "@protected": True,
                "id": "@id",
                "type": "@type",
                "challenge": "https://w3id.org/security#challenge",
                "created": {
                    "@id": "http://purl.org/dc/terms/created",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "domain": "https://w3id.org/security#domain",
                "expires": {
                    "@id": "https://w3id.org/security#expiration",
                    "@type": "http://www.w3.org/2001/XMLSchema#dateTime",
                },
                "nonce": "https://w3id.org/security#nonce",
                "proofPurpose": {
                    "@id": "https://w3id.org/security#proofPurpose",
                    "@type": "@vocab",
                    "@context": {
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "assertionMethod": {
                            "@id": "https://w3id.org/security#assertionMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "authentication": {
                            "@id": "https://w3id.org/security#authenticationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "capabilityInvocation": {
                            "@id": "https://w3id.org/security#capabilityInvocationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "capabilityDelegation": {
                            "@id": "https://w3id.org/security#capabilityDelegationMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                        "keyAgreement": {
                            "@id": "https://w3id.org/security#keyAgreementMethod",
                            "@type": "@id",
                            "@container": "@set",
                        },
                    },
                },
                "proofValue": {
                    "@id": "https://w3id.org/security#proofValue",
                    "@type": "https://w3id.org/security#multibase",
                },
                "verificationMethod": {
                    "@id": "https://w3id.org/security#verificationMethod",
                    "@type": "@id",
                },
            },
        },
    }
}



"""Classes for managing profile information within a request context."""

import logging
from abc import ABC, abstractmethod
from typing import Any, Mapping, Optional, Type
from weakref import ref

from ..config.base import InjectionError
from ..config.injection_context import InjectionContext
from ..config.injector import BaseInjector, InjectType
from ..config.provider import BaseProvider
from ..config.settings import BaseSettings
from ..utils.classloader import ClassLoader, ClassNotFoundError
from .error import ProfileSessionInactiveError
from .event_bus import Event, EventBus

LOGGER = logging.getLogger(__name__)


class Profile(ABC):
    """Base abstraction for handling identity-related state."""

    BACKEND_NAME: Optional[str] = None
    DEFAULT_NAME: str = "default"

    def __init__(
        self,
        *,
        context: Optional[InjectionContext] = None,
        name: Optional[str] = None,
        created: bool = False,
    ):
        """Initialize a base profile."""
        self._created = created
        self._name = name or Profile.DEFAULT_NAME

        context = context or InjectionContext()
        self._context = context.start_scope()
        self._context.injector.bind_instance(Profile, ref(self))

    @property
    def backend(self) -> str:
        """Accessor for the backend implementation name."""
        return self.__class__.BACKEND_NAME

    @property
    def context(self) -> InjectionContext:
        """Accessor for the injection context."""
        return self._context

    @property
    def created(self) -> bool:
        """Accessor for the created flag indicating a new profile."""
        return self._created

    @property
    def name(self) -> str:
        """Accessor for the profile name."""
        return self._name

    @property
    def settings(self) -> BaseSettings:
        """Accessor for scope-specific settings."""
        return self._context.settings

    @abstractmethod
    def session(self, context: Optional[InjectionContext] = None) -> "ProfileSession":
        """Start a new interactive session with no transaction support requested."""

    @abstractmethod
    def transaction(self, context: Optional[InjectionContext] = None) -> "ProfileSession":
        """Start a new interactive session with commit and rollback support.

        If the current backend does not support transactions, then commit
        and rollback operations of the session will not have any effect.
        """

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            cls: The base class to retrieve an instance of
            base_cls: The base class to retrieve
            settings: An optional mapping providing configuration to the provider

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject(base_cls, settings)

    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject_or(base_cls, settings, default)

    async def close(self):
        """Close the profile instance."""

    async def remove(self):
        """Remove the profile."""

    async def notify(self, topic: str, payload: Any):
        """Signal an event."""
        event_bus = self.inject_or(EventBus)
        if event_bus:
            await event_bus.notify(self, Event(topic, payload))

    def __repr__(self) -> str:
        """Get a human readable string."""
        return "<{}(backend={}, name={})>".format(
            self.__class__.__name__, self.backend, self.name
        )

    def __eq__(self, other) -> bool:
        """Equality checks for profiles.

        Multiple profile instances can exist at the same time but point to the
        same profile. This allows us to test equality based on the profile
        pointed to by the instance rather than by object reference comparison.
        """
        if not isinstance(other, Profile):
            return False

        if type(self) is not type(other):
            return False

        return self.name == other.name


class ProfileManager(ABC):
    """Handle provision and open for profile instances."""

    def __init__(self):
        """Initialize the profile manager."""

    @abstractmethod
    async def provision(
        self, context: InjectionContext, config: Mapping[str, Any] = None
    ) -> Profile:
        """Provision a new instance of a profile."""

    @abstractmethod
    async def open(
        self, context: InjectionContext, config: Mapping[str, Any] = None
    ) -> Profile:
        """Open an instance of an existing profile."""


class ProfileSession(ABC):
    """An active connection to the profile management backend."""

    def __init__(
        self,
        profile: Profile,
        *,
        context: Optional[InjectionContext] = None,
        settings: Mapping[str, Any] = None,
    ):
        """Initialize a base profile session."""
        self._active = False
        self._awaited = False
        self._entered = 0
        self._context = (context or profile.context).start_scope(settings)
        self._profile = profile
        self._events = []

        self._context.injector.bind_instance(ProfileSession, ref(self))

    async def _setup(self):
        """Create the underlying session or transaction."""

    async def _teardown(self, commit: Optional[bool] = None):
        """Dispose of the underlying session or transaction."""

    def __await__(self):
        """Coroutine magic method.

        A session must be awaited or used as an async context manager.
        """

        async def _init():
            if not self._active:
                await self._setup()
                self._active = True
            self._awaited = True
            return self

        return _init().__await__()

    async def __aenter__(self):
        """Async context manager entry."""
        if not self._active:
            await self._setup()
            self._active = True
        self._entered += 1
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        self._entered -= 1
        if not self._awaited and not self._entered:
            await self._teardown()
            self._active = False

    @property
    def active(self) -> bool:
        """Accessor for the session active state."""
        return self._active

    @property
    def context(self) -> InjectionContext:
        """Accessor for the associated injection context."""
        return self._context

    @property
    def settings(self) -> BaseSettings:
        """Accessor for scope-specific settings."""
        return self._context.settings

    @property
    def is_transaction(self) -> bool:
        """Check if the session supports commit and rollback operations."""
        return False

    @property
    def profile(self) -> Profile:
        """Accessor for the associated profile instance."""
        return self._profile

    async def commit(self):
        """Commit any updates performed within the transaction.

        If the current session is not a transaction, then nothing is performed.
        """
        if not self._active:
            raise ProfileSessionInactiveError()
        await self._teardown(commit=True)

        # emit any pending events
        for event in self._events:
            await self.emit_event(event["topic"], event["payload"], force_emit=True)
        self._events = []

        self._active = False

    async def rollback(self):
        """Roll back any updates performed within the transaction.

        If the current session is not a transaction, then nothing is performed.
        """
        if not self._active:
            raise ProfileSessionInactiveError()
        await self._teardown(commit=False)

        # clear any pending events
        self._events = []

        self._active = False

    async def emit_event(self, topic: str, payload: Any, force_emit: bool = False):
        """Emit an event.

        If we are in an active transaction, just queue the event, otherwise emit it.

        Args:
            topic (str): The topic of the event.
            payload (Any): The payload of the event.
            force_emit (bool, optional): If True, force the event to be emitted even
                if there is an active transaction. Defaults to False.
        """

        if force_emit or (not self.is_transaction):
            # just emit directly
            await self.profile.notify(topic, payload)
        else:
            # add to queue
            self._events.append(
                {
                    "topic": topic,
                    "payload": payload,
                }
            )

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            base_cls (Type[InjectType]): The base class to retrieve an instance of.
            settings (Mapping[str, object], optional): An optional mapping providing
                configuration to the provider.

        Returns:
            InjectType: An instance of the base class, or None.

        Raises:
            ProfileSessionInactiveError: If the profile session is inactive.

        """
        if not self._active:
            raise ProfileSessionInactiveError()
        return self._context.inject(base_cls, settings)

    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """
        if not self._active:
            raise ProfileSessionInactiveError()
        return self._context.inject_or(base_cls, settings, default)

    def __repr__(self) -> str:
        """Get a human readable string."""
        return "<{}(active={}, is_transaction={})>".format(
            self.__class__.__name__, self._active, self.is_transaction
        )


class ProfileManagerProvider(BaseProvider):
    """The standard profile manager provider which keys off the selected wallet type."""

    MANAGER_TYPES = {
        "askar": "acapy_agent.askar.profile.AskarProfileManager",
        "askar-anoncreds": "acapy_agent.askar.profile_anon.AskarAnonProfileManager",
    }

    def __init__(self):
        """Initialize the profile manager provider."""
        self._inst = {}

    def provide(self, settings: BaseSettings, injector: BaseInjector):
        """Create the profile manager instance."""
        mgr_type = settings.get_value("wallet.type", default="askar")

        # mgr_type may be a fully qualified class name
        mgr_class = self.MANAGER_TYPES.get(mgr_type.lower(), mgr_type)

        if mgr_class not in self._inst:
            LOGGER.info("Create profile manager: %s", mgr_type)
            try:
                self._inst[mgr_class] = ClassLoader.load_class(mgr_class)()
            except ClassNotFoundError as err:
                raise InjectionError(f"Unknown profile manager: {mgr_type}") from err

        return self._inst[mgr_class]



from unittest import IsolatedAsyncioTestCase

from aries_askar import AskarError, AskarErrorCode, Store

from acapy_agent.tests import mock

from ...core.error import ProfileDuplicateError, ProfileError, ProfileNotFoundError
from ..store import AskarOpenStore, AskarStoreConfig


class TestStoreConfig(IsolatedAsyncioTestCase):
    key_derivation_method = "Raw"
    key = "key"
    storage_type = "postgres"

    async def test_init_success(self):
        config = {
            "key_derivation_method": self.key_derivation_method,
            "key": self.key,
            "storage_type": self.storage_type,
        }

        askar_store = AskarStoreConfig(config)

        assert askar_store.key_derivation_method == self.key_derivation_method
        assert askar_store.key == self.key
        assert askar_store.storage_type == self.storage_type


class TestStoreOpen(IsolatedAsyncioTestCase):
    key_derivation_method = "Raw"
    key = "key"
    storage_type = "default"

    @mock.patch.object(Store, "open", autospec=True)
    async def test_open_store(self, mock_store_open):
        config = {
            "key_derivation_method": self.key_derivation_method,
            "key": self.key,
            "storage_type": self.storage_type,
        }

        store = await AskarStoreConfig(config).open_store()
        assert isinstance(store, AskarOpenStore)
        assert mock_store_open.called

    @mock.patch.object(Store, "open")
    async def test_open_store_fails(self, mock_store_open):
        config = {
            "key_derivation_method": self.key_derivation_method,
            "key": self.key,
            "storage_type": self.storage_type,
        }

        mock_store_open.side_effect = [
            AskarError(AskarErrorCode.NOT_FOUND, message="testing"),
            AskarError(AskarErrorCode.DUPLICATE, message="testing"),
            AskarError(AskarErrorCode.ENCRYPTION, message="testing"),
        ]

        with self.assertRaises(ProfileNotFoundError):
            await AskarStoreConfig(config).open_store()
        with self.assertRaises(ProfileDuplicateError):
            await AskarStoreConfig(config).open_store()
        with self.assertRaises(ProfileError):
            await AskarStoreConfig(config).open_store()

    @mock.patch.object(Store, "open")
    @mock.patch.object(Store, "rekey")
    async def test_open_store_fail_retry_with_rekey(self, mock_store_open, mock_rekey):
        config = {
            "key_derivation_method": self.key_derivation_method,
            "key": self.key,
            "storage_type": self.storage_type,
            "rekey": "rekey",
        }

        mock_store_open.side_effect = [
            AskarError(AskarErrorCode.ENCRYPTION, message="testing"),
            mock.AsyncMock(auto_spec=True),
        ]

        store = await AskarStoreConfig(config).open_store()

        assert isinstance(store, AskarOpenStore)
        assert mock_rekey.called

    @mock.patch.object(Store, "open")
    @mock.patch.object(Store, "rekey")
    async def test_open_store_fail_retry_with_rekey_fails(
        self, mock_store_open, mock_rekey
    ):
        config = {
            "key_derivation_method": self.key_derivation_method,
            "key": self.key,
            "storage_type": self.storage_type,
            "rekey": "rekey",
        }

        mock_store_open.side_effect = [
            AskarError(AskarErrorCode.ENCRYPTION, message="testing"),
            mock.AsyncMock(auto_spec=True),
        ]

        store = await AskarStoreConfig(config).open_store()

        assert isinstance(store, AskarOpenStore)
        assert mock_rekey.called



"""Message and inner object type identifiers for present-proof protocol v2.0."""

from ...didcomm_prefix import DIDCommPrefix
from .messages.pres_format import V20PresFormat

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "eace815c3e8598d4a8dd7881d8c731fdb2bcc0aa/features/0454-present-proof-v2"
)

# Message types
PRES_20_PROPOSAL = "present-proof/2.0/propose-presentation"
PRES_20_REQUEST = "present-proof/2.0/request-presentation"
PRES_20 = "present-proof/2.0/presentation"
PRES_20_ACK = "present-proof/2.0/ack"
PRES_20_PROBLEM_REPORT = "present-proof/2.0/problem-report"

PROTOCOL_PACKAGE = "acapy_agent.protocols.present_proof.v2_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        PRES_20_PROPOSAL: f"{PROTOCOL_PACKAGE}.messages.pres_proposal.V20PresProposal",
        PRES_20_REQUEST: f"{PROTOCOL_PACKAGE}.messages.pres_request.V20PresRequest",
        PRES_20: f"{PROTOCOL_PACKAGE}.messages.pres.V20Pres",
        PRES_20_ACK: f"{PROTOCOL_PACKAGE}.messages.pres_ack.V20PresAck",
        PRES_20_PROBLEM_REPORT: (
            f"{PROTOCOL_PACKAGE}.messages.pres_problem_report.V20PresProblemReport"
        ),
    }
)

# Format specifications
ATTACHMENT_FORMAT = {
    PRES_20_PROPOSAL: {
        V20PresFormat.Format.ANONCREDS.api: "anoncreds/proof-proposal@v1.0",
        V20PresFormat.Format.INDY.api: "hlindy/proof-req@v2.0",
        V20PresFormat.Format.DIF.api: "dif/presentation-exchange/definitions@v1.0",
    },
    PRES_20_REQUEST: {
        V20PresFormat.Format.ANONCREDS.api: "anoncreds/proof-request@v1.0",
        V20PresFormat.Format.INDY.api: "hlindy/proof-req@v2.0",
        V20PresFormat.Format.DIF.api: "dif/presentation-exchange/definitions@v1.0",
    },
    PRES_20: {
        V20PresFormat.Format.ANONCREDS.api: "anoncreds/proof@v1.0",
        V20PresFormat.Format.INDY.api: "hlindy/proof@v2.0",
        V20PresFormat.Format.DIF.api: "dif/presentation-exchange/submission@v1.0",
    },
}

CONTROLLERS = DIDCommPrefix.qualify_all(
    {"present-proof/2.0": f"{PROTOCOL_PACKAGE}.controller.Controller"}
)



"""Rotate message."""

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import PROTOCOL_PACKAGE, ROTATE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.rotate_handler.RotateHandler"


class Rotate(AgentMessage):
    """Class representing a rotate message."""

    class Meta:
        """Rotate metadata."""

        handler_class = HANDLER_CLASS
        message_type = ROTATE
        schema_class = "RotateSchema"

    def __init__(self, *, to_did: str, **kwargs):
        """Initialize a Rotate message instance."""
        super().__init__(**kwargs)
        self.to_did = to_did


class RotateSchema(AgentMessageSchema):
    """Schema for Rotate class."""

    class Meta:
        """RotateSchema metadata."""

        model_class = Rotate
        unknown = EXCLUDE

    to_did = fields.Str(
        required=True,
        allow_none=False,
        metadata={
            "description": "The DID the rotating party is rotating to",
            "example": "did:example:newdid",
        },
    )



import pytest

from ..connection_target import ConnectionTarget

TEST_DID_UNQUALIFIED = "55GkHamhTU1ZbTbV2ab9DE"
TEST_DID_SOV = "did:sov:55GkHamhTU1ZbTbV2ab9DE"
TEST_DID_PEER = "did:peer:WgWxqztrNooG92RXvxSTWv"
TEST_DID_WEB = "did:web:example"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_ENDPOINT = "http://localhost"


class TestConnectionTarget:
    @pytest.mark.parametrize(
        "did", [TEST_DID_UNQUALIFIED, TEST_DID_SOV, TEST_DID_PEER, TEST_DID_WEB]
    )
    def test_deser(self, did):
        target = ConnectionTarget(
            did=did,
            endpoint=TEST_ENDPOINT,
            label="a label",
            recipient_keys=[TEST_VERKEY],
            routing_keys=[TEST_VERKEY],
            sender_key=TEST_VERKEY,
        )
        serial = target.serialize()
        serial["extra-stuff"] = "to exclude"
        deser = ConnectionTarget.deserialize(serial)

        assert deser.did == target.did
        assert deser.endpoint == target.endpoint
        assert deser.label == target.label
        assert deser.recipient_keys == target.recipient_keys
        assert deser.routing_keys == target.routing_keys
        assert deser.sender_key == target.sender_key



"""Interfaces and base classes for DID Resolution."""

import logging

from ..config.injection_context import InjectionContext
from ..config.provider import ClassProvider
from ..resolver.did_resolver import DIDResolver
from .base import ResolverError

LOGGER = logging.getLogger(__name__)


async def setup(context: InjectionContext):
    """Set up default resolvers."""
    registry = context.inject_or(DIDResolver)
    if not registry:
        LOGGER.warning("No DID Resolver instance found in context")
        return

    legacy_resolver = ClassProvider(
        "acapy_agent.resolver.default.legacy_peer.LegacyPeerDIDResolver"
    ).provide(context.settings, context.injector)
    await legacy_resolver.setup(context)
    registry.register_resolver(legacy_resolver)

    key_resolver = ClassProvider(
        "acapy_agent.resolver.default.key.KeyDIDResolver"
    ).provide(context.settings, context.injector)
    await key_resolver.setup(context)
    registry.register_resolver(key_resolver)

    jwk_resolver = ClassProvider(
        "acapy_agent.resolver.default.jwk.JwkDIDResolver"
    ).provide(context.settings, context.injector)
    await jwk_resolver.setup(context)
    registry.register_resolver(jwk_resolver)

    if not context.settings.get("ledger.disabled"):
        indy_resolver = ClassProvider(
            "acapy_agent.resolver.default.indy.IndyDIDResolver"
        ).provide(context.settings, context.injector)
        await indy_resolver.setup(context)
        registry.register_resolver(indy_resolver)
    else:
        LOGGER.info("Ledger is not configured, not loading IndyDIDResolver")

    web_resolver = ClassProvider(
        "acapy_agent.resolver.default.web.WebDIDResolver"
    ).provide(context.settings, context.injector)
    await web_resolver.setup(context)
    registry.register_resolver(web_resolver)

    webvh_resolver = ClassProvider(
        "acapy_agent.resolver.default.webvh.WebvhDIDResolver"
    ).provide(context.settings, context.injector)
    await webvh_resolver.setup(context)
    registry.register_resolver(webvh_resolver)

    if context.settings.get("resolver.universal"):
        try:
            universal_resolver = ClassProvider(
                "acapy_agent.resolver.default.universal.UniversalResolver"
            ).provide(context.settings, context.injector)
            await universal_resolver.setup(context)
            registry.register_resolver(universal_resolver)
        except ResolverError as err:
            LOGGER.warning(f"Universal Resolver setup failed: {err}")

    peer_did_1_resolver = ClassProvider(
        "acapy_agent.resolver.default.peer1.PeerDID1Resolver"
    ).provide(context.settings, context.injector)
    await peer_did_1_resolver.setup(context)
    registry.register_resolver(peer_did_1_resolver)

    peer_did_2_resolver = ClassProvider(
        "acapy_agent.resolver.default.peer2.PeerDID2Resolver"
    ).provide(context.settings, context.injector)
    await peer_did_2_resolver.setup(context)
    registry.register_resolver(peer_did_2_resolver)

    peer_did_3_resolver = ClassProvider(
        "acapy_agent.resolver.default.peer3.PeerDID3Resolver"
    ).provide(context.settings, context.injector)
    await peer_did_3_resolver.setup(context)
    registry.register_resolver(peer_did_3_resolver)

    peer_did_4_resolver = ClassProvider(
        "acapy_agent.resolver.default.peer4.PeerDID4Resolver"
    ).provide(context.settings, context.injector)
    await peer_did_4_resolver.setup(context)
    registry.register_resolver(peer_did_4_resolver)



from unittest import IsolatedAsyncioTestCase

from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.askar import AskarWallet
from .....wallet.key_type import ED25519
from ...error import LinkedDataProofException
from ..wallet_key_pair import WalletKeyPair


class TestWalletKeyPair(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

    async def test_sign_x_no_public_key(self):
        key_pair = WalletKeyPair(profile=self.profile, key_type=ED25519)

        with self.assertRaises(LinkedDataProofException) as context:
            await key_pair.sign(b"Message")
        assert "No key to sign with" in str(context.exception)

    async def test_sign(self):
        public_key_base58 = "verkey"
        key_pair = WalletKeyPair(
            profile=self.profile,
            key_type=ED25519,
            public_key_base58=public_key_base58,
        )
        signed = mock.MagicMock()

        with mock.patch.object(
            AskarWallet,
            "sign_message",
            mock.CoroutineMock(return_value=signed),
        ) as sign_message:
            singed_ret = await key_pair.sign(b"Message")

            assert signed == singed_ret
            sign_message.assert_called_once_with(
                message=b"Message", from_verkey=public_key_base58
            )

    async def test_verify_x_no_public_key(self):
        key_pair = WalletKeyPair(profile=self.profile, key_type=ED25519)

        with self.assertRaises(LinkedDataProofException) as context:
            await key_pair.verify(b"Message", b"signature")
        assert "No key to verify with" in str(context.exception)

    async def test_verify(self):
        public_key_base58 = "verkey"
        key_pair = WalletKeyPair(
            profile=self.profile,
            key_type=ED25519,
            public_key_base58=public_key_base58,
        )

        with mock.patch.object(
            AskarWallet,
            "verify_message",
            mock.CoroutineMock(return_value=True),
        ) as verify_message:
            verified = await key_pair.verify(b"Message", b"signature")

            assert verified
            verify_message.assert_called_once_with(
                message=b"Message",
                signature=b"signature",
                from_verkey=public_key_base58,
                key_type=ED25519,
            )

    async def test_from_verification_method_x_no_public_key_base58(self):
        key_pair = WalletKeyPair(profile=self.profile, key_type=ED25519)

        with self.assertRaises(LinkedDataProofException) as context:
            key_pair.from_verification_method({})
        assert "Unrecognized" in str(context.exception)



from unittest import IsolatedAsyncioTestCase

from acapy_agent.tests import mock

from ...config.injection_context import InjectionContext
from ...utils.classloader import ClassLoader, DeferLoad
from ..protocol_registry import ProtocolRegistry


class TestProtocolRegistry(IsolatedAsyncioTestCase):
    no_type_message = {"a": "b"}
    unknown_type_message = {"@type": 1}
    test_message_type = "doc/protocol/1.0/message"
    test_protocol = "doc/protocol/1.0"
    test_protocol_queries = ["*", "doc/protocol/1.0", "doc/proto*"]
    test_protocol_queries_fail = ["", "nomatch", "nomatch*"]
    test_message_cls = "fake_msg_cls"
    test_controller = "fake_controller"

    def setUp(self):
        self.registry = ProtocolRegistry()

    def test_protocols(self):
        self.registry.register_message_types(
            {self.test_message_type: self.test_message_cls}
        )
        self.registry.register_controllers({self.test_message_type: self.test_controller})

        assert list(self.registry.message_types) == [self.test_message_type]
        assert list(self.registry.protocols) == [self.test_protocol]

    def test_message_type_query(self):
        self.registry.register_message_types(
            {self.test_message_type: self.test_message_cls}
        )
        for q in self.test_protocol_queries:
            matches = self.registry.protocols_matching_query(q)
            assert tuple(matches) == (self.test_protocol,)
        for q in self.test_protocol_queries_fail:
            matches = self.registry.protocols_matching_query(q)
            assert matches == ()

    def test_registration_with_minor_version(self):
        MSG_PATH = "acapy_agent.protocols.introduction.v0_1.messages"
        test_typesets = {
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/introduction-service/1.0/fake-forward-invitation": f"{MSG_PATH}.forward_invitation.ForwardInvitation",
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/introduction-service/1.0/fake-invitation": f"{MSG_PATH}.invitation.Invitation",
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/introduction-service/1.0/fake-invitation-request": f"{MSG_PATH}.invitation_request.InvitationRequest",
            "https://didcom.org/introduction-service/1.0/fake-forward-invitation": f"{MSG_PATH}.forward_invitation.ForwardInvitation",
            "https://didcom.org/introduction-service/1.0/fake-invitation": f"{MSG_PATH}.invitation.Invitation",
            "https://didcom.org/introduction-service/1.0/fake-invitation-request": f"{MSG_PATH}.invitation_request.InvitationRequest",
        }
        test_version_def = {
            "current_minor_version": 0,
            "major_version": 1,
            "minimum_minor_version": 0,
            "path": "v0_1",
        }
        self.registry.register_message_types(test_typesets, test_version_def)
        assert (
            "https://didcom.org/introduction-service/1.0/fake-forward-invitation"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/introduction-service/1.0/fake-invitation"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/introduction-service/1.0/fake-invitation-request"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/introduction-service/1.0/fake-forward-invitation"
            in self.registry.message_types
        )

    def test_register_msg_types_for_multiple_minor_versions(self):
        MSG_PATH = "acapy_agent.protocols.out_of_band.v1_0.messages"
        test_typesets = {
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/invitation": f"{MSG_PATH}.invitation.Invitation",
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/handshake-reuse": f"{MSG_PATH}.reuse.HandshakeReuse",
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/handshake-reuse-accepted": f"{MSG_PATH}.reuse_accept.HandshakeReuseAccept",
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/problem_report": f"{MSG_PATH}.problem_report.OOBProblemReport",
            "https://didcom.org/out-of-band/1.1/invitation": f"{MSG_PATH}.invitation.Invitation",
            "https://didcom.org/out-of-band/1.1/handshake-reuse": f"{MSG_PATH}.reuse.HandshakeReuse",
            "https://didcom.org/out-of-band/1.1/handshake-reuse-accepted": f"{MSG_PATH}.reuse_accept.HandshakeReuseAccept",
            "https://didcom.org/out-of-band/1.1/problem_report": f"{MSG_PATH}.problem_report.OOBProblemReport",
        }
        test_version_def = {
            "current_minor_version": 1,
            "major_version": 1,
            "minimum_minor_version": 0,
            "path": "v0_1",
        }
        self.registry.register_message_types(test_typesets, test_version_def)
        assert (
            "https://didcom.org/out-of-band/1.0/invitation" in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.0/handshake-reuse"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.0/handshake-reuse-accepted"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.0/problem_report"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.1/invitation" in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.1/handshake-reuse"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.1/handshake-reuse-accepted"
            in self.registry.message_types
        )
        assert (
            "https://didcom.org/out-of-band/1.1/problem_report"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.0/invitation"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.0/handshake-reuse"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.0/handshake-reuse-accepted"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.0/problem_report"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/invitation"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/handshake-reuse"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/handshake-reuse-accepted"
            in self.registry.message_types
        )
        assert (
            "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/out-of-band/1.1/problem_report"
            in self.registry.message_types
        )

    async def test_disclosed(self):
        self.registry.register_message_types(
            {self.test_message_type: self.test_message_cls}
        )
        mocked = mock.MagicMock()
        mocked.return_value.check_access = mock.CoroutineMock()
        mocked.return_value.check_access.return_value = True
        mocked.return_value.determine_roles = mock.CoroutineMock()
        mocked.return_value.determine_roles.return_value = ["ROLE"]
        self.registry.register_controllers({self.test_protocol: mocked})
        protocols = [self.test_protocol]
        ctx = InjectionContext()
        published = await self.registry.prepare_disclosed(ctx, protocols)
        mocked.return_value.check_access.assert_called_once_with(ctx)
        mocked.return_value.determine_roles.assert_called_once_with(ctx)
        assert len(published) == 1
        assert published[0]["pid"] == self.test_protocol
        assert published[0]["roles"] == ["ROLE"]

    async def test_disclosed_str(self):
        self.registry.register_message_types(
            {self.test_message_type: self.test_message_cls}
        )
        self.registry.register_controllers({self.test_protocol: "mock-class-name"})
        protocols = [self.test_protocol]
        ctx = InjectionContext()

        class Mockery:
            def __init__(self, protocol):
                self.protocol = protocol

            async def check_access(self, context):
                return False

        with mock.patch.object(ClassLoader, "load_class", mock.MagicMock()) as load_class:
            load_class.return_value = Mockery
            published = await self.registry.prepare_disclosed(ctx, protocols)
            assert not published

    def test_resolve_message_class_str(self):
        self.registry.register_message_types(
            {self.test_message_type: self.test_message_cls}
        )
        result = self.registry.resolve_message_class(self.test_message_type)
        assert isinstance(result, DeferLoad)
        assert result._cls_path == self.test_message_cls

    def test_resolve_message_class_no_major_version_support(self):
        result = self.registry.resolve_message_class("doc/proto/1.2/hello")
        assert result is None

    def test_resolve_message_load_class_str(self):
        message_type_a = "doc/proto/1.2/aaa"
        self.registry.register_message_types(
            {message_type_a: self.test_message_cls},
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 2,
                "path": "v1_2",
            },
        )
        result = self.registry.resolve_message_class("doc/proto/1.1/aaa")
        assert isinstance(result, DeferLoad)
        assert result._cls_path == self.test_message_cls

    def test_resolve_message_load_class_none(self):
        message_type_a = "doc/proto/1.2/aaa"
        self.registry.register_message_types(
            {message_type_a: self.test_message_cls},
            version_definition={
                "major_version": 1,
                "minimum_minor_version": 0,
                "current_minor_version": 2,
                "path": "v1_2",
            },
        )
        result = self.registry.resolve_message_class("doc/proto/1.2/bbb")
        assert result is None

    def test_repr(self):
        assert isinstance(repr(self.registry), str)



import json
from unittest import IsolatedAsyncioTestCase

from aries_askar import AskarErrorCode

from ...admin.request_context import AdminRequestContext
from ...anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ...indy.holder import IndyHolder
from ...ledger.base import BaseLedger
from ...storage.vc_holder.base import VCHolder
from ...storage.vc_holder.vc_record import VCRecord
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import routes as test_module

VC_RECORD = VCRecord(
    contexts=[
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
    ],
    expanded_types=[
        "https://www.w3.org/2018/credentials#VerifiableCredential",
        "https://example.org/examples#UniversityDegreeCredential",
    ],
    issuer_id="https://example.edu/issuers/565049",
    subject_ids=["did:example:ebfeb1f712ebc6f1c276e12ec21"],
    proof_types=["Ed25519Signature2018"],
    schema_ids=["https://example.org/examples/degree.json"],
    cred_value={"...": "..."},
    given_id="http://example.edu/credentials/3732",
    cred_tags={"some": "tag"},
)


class TestHolderRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = self.profile.context
        setattr(self.context, "profile", self.profile)

        self.request_dict = {"context": self.context}
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )
        self.profile.context.injector.bind_instance(
            IndyHolder, mock.MagicMock(IndyHolder, autospec=True)
        )

    async def test_credentials_get(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps({"hello": "world"})
        )
        self.profile.context.injector.bind_instance(
            IndyHolder,
            mock_holder,
        )

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_get(self.request)
            json_response.assert_called_once_with({"hello": "world"})
            assert result is json_response.return_value

    @mock.patch.object(AnonCredsHolder, "get_credential")
    async def test_credentials_get_with_anoncreds(self, mock_get_credential):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            match_info={"credential_id": "dummy"},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential=mock.CoroutineMock(return_value="test-credential")
            )
        )

        mock_get_credential.return_value = json.dumps({"hello": "world"})

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_get(self.request)
            json_response.assert_called_once_with({"hello": "world"})
            assert result is json_response.return_value
            assert mock_get_credential.called

    async def test_credentials_get_not_found(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credential = mock.CoroutineMock(
            side_effect=test_module.WalletNotFoundError()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.credentials_get(self.request)

    async def test_credentials_revoked(self):
        self.request.match_info = {"credential_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseLedger, mock.create_autospec(BaseLedger)
        )
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.credential_revoked = mock.CoroutineMock(return_value=False)
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_revoked(self.request)
            json_response.assert_called_once_with({"revoked": False})
            assert result is json_response.return_value

    @mock.patch.object(AnonCredsHolder, "credential_revoked")
    async def test_credentials_revoked_with_anoncreds(self, mock_credential_revoked):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            match_info={"credential_id": "dummy"},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential=mock.CoroutineMock(return_value="test-credential")
            )
        )

        mock_credential_revoked.return_value = False

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_revoked(self.request)
            json_response.assert_called_once_with({"revoked": False})
            assert result is json_response.return_value

    async def test_credentials_revoked_no_ledger(self):
        self.request.match_info = {"credential_id": "dummy"}

        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.credentials_revoked(self.request)

    async def test_credentials_not_found(self):
        self.request.match_info = {"credential_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseLedger, mock.create_autospec(BaseLedger)
        )

        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.credential_revoked = mock.CoroutineMock(
            side_effect=test_module.WalletNotFoundError("no such cred")
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.credentials_revoked(self.request)

    async def test_credentials_x_ledger(self):
        self.request.match_info = {"credential_id": "dummy"}
        self.profile.context.injector.bind_instance(
            BaseLedger, mock.create_autospec(BaseLedger)
        )
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.credential_revoked = mock.CoroutineMock(
            side_effect=test_module.LedgerError("down for maintenance")
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credentials_revoked(self.request)

    async def test_attribute_mime_types_get(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_mime_type = mock.CoroutineMock(
            side_effect=[None, {"a": "application/jpeg"}]
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            await test_module.credentials_attr_mime_types_get(self.request)
            mock_response.assert_called_once_with({"results": None})

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            await test_module.credentials_attr_mime_types_get(self.request)
            mock_response.assert_called_once_with({"results": {"a": "application/jpeg"}})

    @mock.patch.object(AnonCredsHolder, "get_mime_type")
    async def test_attribute_mime_types_get_with_anoncreds(self, mock_get_mime_type):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            match_info={"credential_id": "dummy"},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential=mock.CoroutineMock(return_value="test-credential")
            )
        )

        mock_get_mime_type.side_effect = [None, {"a": "application/jpeg"}]

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            await test_module.credentials_attr_mime_types_get(self.request)
            mock_response.assert_called_once_with({"results": None})

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            await test_module.credentials_attr_mime_types_get(self.request)
            mock_response.assert_called_once_with({"results": {"a": "application/jpeg"}})
            assert mock_get_mime_type.called

    async def test_credentials_remove(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.delete_credential = mock.CoroutineMock(return_value=None)
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_remove(self.request)
            json_response.assert_called_once_with({})
            assert result is json_response.return_value

    @mock.patch.object(AnonCredsHolder, "delete_credential")
    async def test_credentials_remove_with_anoncreds(self, mock_delete_credential):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            match_info={"credential_id": "dummy"},
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential=mock.CoroutineMock(return_value="test-credential")
            )
        )

        # AnonCreds holder errors
        mock_delete_credential.side_effect = [
            None,
            AnonCredsHolderError("anoncreds error", error_code=AskarErrorCode.NOT_FOUND),
            AnonCredsHolderError("anoncreds error", error_code=AskarErrorCode.UNEXPECTED),
            AnonCredsHolderError("anoncreds error", error_code=AskarErrorCode.NOT_FOUND),
        ]

        # Indy holder errors
        mock_indy_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_indy_holder.delete_credential.side_effect = [
            test_module.WalletNotFoundError(),  # Indy not found after anoncreds not found
            None,  # Indy found after second anoncreds not found side effect
        ]
        self.profile.context.injector.bind_instance(IndyHolder, mock_indy_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            # First mock delete has no side effect; delete succeeds
            result = await test_module.credentials_remove(self.request)
            json_response.assert_called_once_with({})
            assert result is json_response.return_value
            assert mock_delete_credential.called

            # Not found after anoncreds not found and indy not found
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credentials_remove(self.request)

            # Bad request after anoncreds unexpected error
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credentials_remove(self.request)

            # Indy found after anoncreds not found
            result = await test_module.credentials_remove(self.request)
            assert result is json_response.return_value

    async def test_credentials_remove_not_found(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.delete_credential = mock.CoroutineMock(
            side_effect=test_module.WalletNotFoundError()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.credentials_remove(self.request)

    async def test_credentials_list(self):
        self.request.query = {"start": "0", "count": "10"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials = mock.CoroutineMock(
            return_value=[{"hello": "world"}]
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_list(self.request)
            json_response.assert_called_once_with({"results": [{"hello": "world"}]})
            assert result is json_response.return_value

    @mock.patch.object(AnonCredsHolder, "get_credentials")
    async def test_credentials_list_with_anoncreds(self, mock_get_credentials):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "wallet.type": "askar-anoncreds",
                "admin.admin_api_key": "secret-key",
            },
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
        }
        self.request = mock.MagicMock(
            match_info={"credential_id": "dummy"},
            query={
                "start": "0",
                "count": "10",
            },
            __getitem__=lambda _, k: self.request_dict[k],
            context=self.context,
            headers={"x-api-key": "secret-key"},
        )
        self.context.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential=mock.CoroutineMock(return_value="test-credential")
            )
        )

        mock_get_credentials.return_value = [{"hello": "world"}]

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.credentials_list(self.request)
            json_response.assert_called_once_with({"results": [{"hello": "world"}]})
            assert result is json_response.return_value

    async def test_credentials_list_x_holder(self):
        self.request.query = {"start": "0", "count": "10"}
        mock_holder = mock.MagicMock(IndyHolder, autospec=True)
        mock_holder.get_credentials = mock.CoroutineMock(
            side_effect=test_module.IndyHolderError()
        )
        self.profile.context.injector.bind_instance(IndyHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credentials_list(self.request)

    async def test_w3c_cred_get(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.retrieve_credential_by_id = mock.CoroutineMock(return_value=VC_RECORD)
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            await test_module.w3c_cred_get(self.request)
            json_response.assert_called_once_with(VC_RECORD.serialize())

    async def test_w3c_cred_get_not_found_x(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.retrieve_credential_by_id = mock.CoroutineMock(
            side_effect=test_module.StorageNotFoundError()
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.w3c_cred_get(self.request)

    async def test_w3c_cred_get_storage_x(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.retrieve_credential_by_id = mock.CoroutineMock(
            side_effect=test_module.StorageError()
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.w3c_cred_get(self.request)

    async def test_w3c_cred_remove(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.retrieve_credential_by_id = mock.CoroutineMock(return_value=VC_RECORD)
        mock_holder.delete_credential = mock.CoroutineMock(return_value=None)
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            result = await test_module.w3c_cred_remove(self.request)
            json_response.assert_called_once_with({})
            assert result is json_response.return_value

    async def test_w3c_cred_remove_not_found_x(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.retrieve_credential_by_id = mock.CoroutineMock(
            side_effect=test_module.StorageNotFoundError()
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.w3c_cred_remove(self.request)

    async def test_w3c_cred_remove_storage_x(self):
        self.request.match_info = {"credential_id": "dummy"}
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.retrieve_credential_by_id = mock.CoroutineMock(return_value=VC_RECORD)
        mock_holder.delete_credential = mock.CoroutineMock(
            side_effect=test_module.StorageError()
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.w3c_cred_remove(self.request)

    async def test_w3c_creds_list(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "types": [
                    "VerifiableCredential",
                    "AlumniCredential",
                ],
                "issuer_id": "https://example.edu/issuers/565049",
                "subject_id": "did:example:ebfeb1f712ebc6f1c276e12ec21",
                "max_results": "1",
            }
        )
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(return_value=[VC_RECORD])
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with mock.patch.object(
            test_module.web, "json_response", mock.Mock()
        ) as json_response:
            await test_module.w3c_creds_list(self.request)
            json_response.assert_called_once_with({"results": [VC_RECORD.serialize()]})

    async def test_w3c_creds_list_not_found_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "types": [
                    "VerifiableCredential",
                    "AlumniCredential",
                ],
                "issuer_id": "https://example.edu/issuers/565049",
                "subject_id": "did:example:ebfeb1f712ebc6f1c276e12ec21",
                "max_results": "1",
            }
        )
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(side_effect=test_module.StorageNotFoundError())
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPNotFound):
            await test_module.w3c_creds_list(self.request)

    async def test_w3c_creds_list_storage_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "types": [
                    "VerifiableCredential",
                    "AlumniCredential",
                ],
                "issuer_id": "https://example.edu/issuers/565049",
                "subject_id": "did:example:ebfeb1f712ebc6f1c276e12ec21",
                "max_results": "1",
            }
        )
        mock_holder = mock.MagicMock(VCHolder, autospec=True)
        mock_holder.search_credentials = mock.MagicMock(
            return_value=mock.MagicMock(
                fetch=mock.CoroutineMock(side_effect=test_module.StorageError())
            )
        )
        self.profile.context.injector.bind_instance(VCHolder, mock_holder)

        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.w3c_creds_list(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""Handler for revoke message."""

from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from ..messages.revoke import Revoke


class RevokeHandler(BaseHandler):
    """Handler for revoke message."""

    RECEIVED_TOPIC = "acapy::revocation-notification-v2::received"
    WEBHOOK_TOPIC = "acapy::webhook::revocation-notification-v2"

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle revoke message."""
        assert isinstance(context.message, Revoke)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        self._logger.debug(
            "Received notification of revocation for %s cred %s with comment: %s",
            context.message.revocation_format,
            context.message.credential_id,
            context.message.comment,
        )
        # Emit a webhook
        if context.settings.get("revocation.monitor_notification"):
            await context.profile.notify(
                self.WEBHOOK_TOPIC,
                {
                    "revocation_format": context.message.revocation_format,
                    "credential_id": context.message.credential_id,
                    "comment": context.message.comment,
                },
            )

        # Emit an event
        await context.profile.notify(
            self.RECEIVED_TOPIC,
            {
                "revocation_format": context.message.revocation_format,
                "credential_id": context.message.credential_id,
                "comment": context.message.comment,
            },
        )



"""Test routes.py"""

import pytest
import pytest_asyncio

from .....config.settings import Settings
from .....core.event_bus import Event, MockEventBus
from .....core.profile import Profile
from .....messaging.responder import BaseResponder, MockResponder
from .....protocols.revocation_notification.v1_0.models.rev_notification_record import (
    RevNotificationRecord,
)
from .....revocation.util import (
    REVOCATION_CLEAR_PENDING_EVENT,
    REVOCATION_EVENT_PREFIX,
    REVOCATION_PUBLISHED_EVENT,
)
from .....storage.error import StorageError, StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import routes as test_module


@pytest.fixture
def responder():
    yield MockResponder()


@pytest_asyncio.fixture
async def profile(responder):
    profile = await create_test_profile()
    profile.context.injector.bind_instance(BaseResponder, responder)
    yield profile


def test_register_events():
    """Test handlers are added on register.

    This test need not be particularly in depth to keep it from getting brittle.
    """
    event_bus = MockEventBus()
    test_module.register_events(event_bus)
    assert event_bus.topic_patterns_to_subscribers


@pytest.mark.asyncio
async def test_on_revocation_published(profile: Profile, responder: MockResponder):
    """Test revocation published event handler."""
    mock_rec = mock.MagicMock()
    mock_rec.cred_rev_id = "mock"
    mock_rec.delete_record = mock.CoroutineMock()
    mock_rec.query_by_rev_reg_id = mock.CoroutineMock(return_value=[mock_rec])

    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_PUBLISHED_EVENT}::mock"
    event = Event(topic, {"rev_reg_id": "mock", "crids": ["mock"]})

    assert isinstance(profile.settings, Settings)

    profile.settings.set_value("revocation.notify", True)

    with mock.patch.object(test_module, "RevNotificationRecord", mock_rec):
        await test_module.on_revocation_published(profile, event)

    mock_rec.query_by_rev_reg_id.assert_called_once()
    mock_rec.delete_record.assert_called_once()
    assert responder.messages

    # Test with integer crids
    mock_rec.cred_rev_id = "1"
    mock_rec.query_by_rev_reg_id = mock.CoroutineMock(return_value=[mock_rec])
    event = Event(topic, {"rev_reg_id": "mock", "crids": [1]})

    with mock.patch.object(test_module, "RevNotificationRecord", mock_rec):
        await test_module.on_revocation_published(profile, event)

    mock_rec.query_by_rev_reg_id.assert_called_once()
    assert mock_rec.delete_record.call_count == 2

    # Test with empty crids
    mock_rec.cred_rev_id = "1"
    mock_rec.query_by_rev_reg_id = mock.CoroutineMock(return_value=[mock_rec])
    event = Event(topic, {"rev_reg_id": "mock", "crids": []})

    with mock.patch.object(test_module, "RevNotificationRecord", mock_rec):
        await test_module.on_revocation_published(profile, event)

    mock_rec.query_by_rev_reg_id.assert_called_once()
    assert mock_rec.delete_record.call_count == 2


@pytest.mark.asyncio
async def test_on_revocation_published_x_not_found(
    profile: Profile, responder: MockResponder
):
    """Test revocation published event handler."""
    mock_rec = mock.MagicMock()
    mock_rec.query_by_rev_reg_id = mock.CoroutineMock(side_effect=StorageNotFoundError)

    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_PUBLISHED_EVENT}::mock"
    event = Event(topic, {"rev_reg_id": "mock", "crids": ["mock"]})

    with mock.patch.object(test_module, "RevNotificationRecord", mock_rec):
        await test_module.on_revocation_published(profile, event)

    mock_rec.query_by_rev_reg_id.assert_called_once()
    assert not responder.messages


@pytest.mark.asyncio
async def test_on_revocation_published_x_storage_error(
    profile: Profile, responder: MockResponder
):
    """Test revocation published event handler."""
    mock_rec = mock.MagicMock()
    mock_rec.query_by_rev_reg_id = mock.CoroutineMock(side_effect=StorageError)

    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_PUBLISHED_EVENT}::mock"
    event = Event(topic, {"rev_reg_id": "mock", "crids": ["mock"]})

    with mock.patch.object(test_module, "RevNotificationRecord", mock_rec):
        await test_module.on_revocation_published(profile, event)

    mock_rec.query_by_rev_reg_id.assert_called_once()
    assert not responder.messages


@pytest.mark.asyncio
async def test_on_pending_cleared(profile: Profile):
    """Test pending revocation cleared event."""
    mock_rec = mock.MagicMock(RevNotificationRecord, autospec=True)
    mock_rec.delete_record = mock.CoroutineMock()
    mock_rec.query_by_rev_reg_id = mock.CoroutineMock(return_value=[mock_rec])

    topic = f"{REVOCATION_EVENT_PREFIX}{REVOCATION_CLEAR_PENDING_EVENT}::mock"
    event = Event(topic, {"rev_reg_id": "mock"})

    with mock.patch.object(test_module, "RevNotificationRecord", mock_rec):
        await test_module.on_pending_cleared(profile, event)

    mock_rec.delete_record.assert_called_once()



"""Test MediationManager."""

from typing import AsyncIterable, Iterable

import pytest
import pytest_asyncio

from .....core.event_bus import EventBus
from .....core.profile import Profile, ProfileSession
from .....did.did_key import DIDKey
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .....wallet.did_method import DIDMethods
from .....wallet.key_type import KeyTypes
from ....routing.v1_0.models.route_record import RouteRecord
from .. import manager as test_module
from ..manager import (
    MediationAlreadyExists,
    MediationManager,
    MediationManagerError,
    MediationNotGrantedError,
)
from ..messages.inner.keylist_update_rule import KeylistUpdateRule
from ..messages.inner.keylist_updated import KeylistUpdated
from ..messages.mediate_deny import MediationDeny
from ..messages.mediate_grant import MediationGrant
from ..messages.mediate_request import MediationRequest
from ..models.mediation_record import MediationRecord

TEST_CONN_ID = "conn-id"
TEST_THREAD_ID = "thread-id"
TEST_ENDPOINT = "https://example.com"
TEST_BASE58_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_VERKEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_ROUTE_RECORD_VERKEY = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"
TEST_ROUTE_VERKEY = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya#z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"

pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture
async def profile():
    """Fixture for profile used in tests."""
    profile = await create_test_profile()
    profile.context.injector.bind_instance(DIDMethods, DIDMethods())
    profile.context.injector.bind_instance(KeyTypes, KeyTypes())
    yield profile


@pytest.fixture
def mock_event_bus(profile: Profile):
    yield profile.inject(EventBus)


@pytest_asyncio.fixture
async def session(profile) -> AsyncIterable[ProfileSession]:  # pylint: disable=W0621
    """Fixture for profile session used in tests."""
    async with profile.session() as session:
        yield session


@pytest.fixture
def manager(profile) -> Iterable[MediationManager]:  # pylint: disable=W0621
    """Fixture for manager used in tests."""
    yield MediationManager(profile)


@pytest.fixture
def record() -> Iterable[MediationRecord]:
    """Fixture for record used in tests."""
    yield MediationRecord(state=MediationRecord.STATE_GRANTED, connection_id=TEST_CONN_ID)


class TestMediationManager:  # pylint: disable=R0904,W0621
    """Test MediationManager."""

    async def test_create_manager_no_profile(self):
        """test_create_manager_no_profile."""
        with pytest.raises(MediationManagerError):
            MediationManager(None)

    async def test_create_did(self, manager, session):
        """test_create_did."""
        # pylint: disable=W0212
        await manager._create_routing_did(session)
        assert await manager._retrieve_routing_did(session)

    async def test_retrieve_did_when_absent(self, manager, session):
        """test_retrieve_did_when_absent."""
        # pylint: disable=W0212
        assert await manager._retrieve_routing_did(session) is None

    async def test_receive_request_no_terms(self, manager):
        """test_receive_request_no_terms."""
        request = MediationRequest()
        record = await manager.receive_request(TEST_CONN_ID, request)
        assert record.connection_id == TEST_CONN_ID

    async def test_receive_request_record_exists(self, session, manager):
        """test_receive_request_no_terms."""
        request = MediationRequest()
        await MediationRecord(connection_id=TEST_CONN_ID).save(session)
        with pytest.raises(MediationAlreadyExists):
            await manager.receive_request(TEST_CONN_ID, request)

    @pytest.mark.skip(
        reason="mediator and recipient terms are only loosely defined in RFC 0211"
    )
    async def test_receive_request_unacceptable_terms(self):
        """test_receive_request_unacceptable_terms."""

    async def test_grant_request(self, session, manager):
        """test_grant_request."""
        # pylint: disable=W0212
        request = MediationRequest()
        record = await manager.receive_request(TEST_CONN_ID, request)
        assert record.connection_id == TEST_CONN_ID
        record, grant = await manager.grant_request(record.mediation_id)
        assert grant.endpoint == session.settings.get("default_endpoint")
        routing_key = await manager._retrieve_routing_did(session)
        routing_key = DIDKey.from_public_key_b58(
            routing_key.verkey, routing_key.key_type
        ).key_id
        assert grant.routing_keys == [routing_key]

    async def test_deny_request(self, manager):
        """test_deny_request."""
        request = MediationRequest()
        record = await manager.receive_request(TEST_CONN_ID, request)
        assert record.connection_id == TEST_CONN_ID
        record, _ = await manager.deny_request(record.mediation_id)

    async def test_update_keylist_delete(self, session, manager, record):
        """test_update_keylist_delete."""
        await RouteRecord(
            connection_id=TEST_CONN_ID, recipient_key=TEST_BASE58_VERKEY
        ).save(session)
        response = await manager.update_keylist(
            record=record,
            updates=[
                KeylistUpdateRule(
                    recipient_key=TEST_VERKEY, action=KeylistUpdateRule.RULE_REMOVE
                )
            ],
        )
        results = response.updated
        assert len(results) == 1
        assert results[0].recipient_key == TEST_VERKEY
        assert results[0].action == KeylistUpdateRule.RULE_REMOVE
        assert results[0].result == KeylistUpdated.RESULT_SUCCESS

    async def test_update_keylist_create(self, manager, record):
        """test_update_keylist_create."""
        response = await manager.update_keylist(
            record=record,
            updates=[
                KeylistUpdateRule(
                    recipient_key=TEST_VERKEY, action=KeylistUpdateRule.RULE_ADD
                )
            ],
        )
        results = response.updated
        assert len(results) == 1
        assert results[0].recipient_key == TEST_VERKEY
        assert results[0].action == KeylistUpdateRule.RULE_ADD
        assert results[0].result == KeylistUpdated.RESULT_SUCCESS

    async def test_update_keylist_create_existing(self, session, manager, record):
        """test_update_keylist_create_existing."""
        await RouteRecord(
            connection_id=TEST_CONN_ID, recipient_key=TEST_BASE58_VERKEY
        ).save(session)
        response = await manager.update_keylist(
            record=record,
            updates=[
                KeylistUpdateRule(
                    recipient_key=TEST_VERKEY, action=KeylistUpdateRule.RULE_ADD
                )
            ],
        )
        results = response.updated
        assert len(results) == 1
        assert results[0].recipient_key == TEST_VERKEY
        assert results[0].action == KeylistUpdateRule.RULE_ADD
        assert results[0].result == KeylistUpdated.RESULT_NO_CHANGE

    async def test_update_keylist_x_not_granted(
        self, manager: MediationManager, record: MediationRecord
    ):
        record.state = MediationRecord.STATE_DENIED
        with pytest.raises(MediationNotGrantedError):
            await manager.update_keylist(record, [])

    async def test_get_keylist(self, session, manager, record):
        """test_get_keylist."""
        await RouteRecord(connection_id=TEST_CONN_ID, recipient_key=TEST_VERKEY).save(
            session
        )
        # Non-server route for verifying filtering
        await RouteRecord(
            role=RouteRecord.ROLE_CLIENT,
            connection_id=TEST_CONN_ID,
            recipient_key=TEST_ROUTE_VERKEY,
        ).save(session)
        results = await manager.get_keylist(record)
        assert len(results) == 1
        assert results[0].connection_id == TEST_CONN_ID
        assert results[0].recipient_key == TEST_VERKEY

    async def test_get_keylist_no_granted_record(self, manager):
        """test_get_keylist_no_granted_record."""
        record = MediationRecord()
        with pytest.raises(MediationNotGrantedError):
            await manager.get_keylist(record)

    async def test_create_keylist_query_response(self, session, manager, record):
        """test_create_keylist_query_response."""
        await RouteRecord(connection_id=TEST_CONN_ID, recipient_key=TEST_VERKEY).save(
            session
        )
        results = await manager.get_keylist(record)
        response = await manager.create_keylist_query_response(results)
        assert len(response.keys) == 1
        assert response.keys[0].recipient_key
        response = await manager.create_keylist_query_response([])
        assert not response.keys

    async def test_get_set_get_default_mediator(
        self,
        session: ProfileSession,
        manager: MediationManager,
        record: MediationRecord,
    ):
        await record.save(session)
        assert await manager.get_default_mediator() is None
        await manager.set_default_mediator(record)
        assert await manager.get_default_mediator() == record

    async def test_set_get_default_mediator_by_id(
        self, manager: MediationManager, session
    ):
        await manager._set_default_mediator_id("test", session)
        assert await manager.get_default_mediator_id() == "test"

    async def test_set_set_get_default_mediator_id(
        self, manager: MediationManager, session
    ):
        await manager._set_default_mediator_id("test", session)
        await manager._set_default_mediator_id("updated", session)
        assert await manager.get_default_mediator_id() == "updated"

    async def test_set_default_mediator_by_id(self, manager: MediationManager):
        with mock.patch.object(
            test_module.MediationRecord, "retrieve_by_id", mock.CoroutineMock()
        ):
            await manager.set_default_mediator_by_id("test")

    async def test_clear_default_mediator(self, manager: MediationManager, session):
        await manager._set_default_mediator_id("test", session)
        assert await manager.get_default_mediator_id()
        await manager.clear_default_mediator()
        assert not await manager.get_default_mediator_id()

    async def test_clear_default_mediator_no_default_set(self, manager: MediationManager):
        await manager.clear_default_mediator()

    async def test_prepare_request(self, manager):
        """test_prepare_request."""
        record, request = await manager.prepare_request(TEST_CONN_ID)
        assert record.connection_id == TEST_CONN_ID
        assert request

    async def test_request_granted_base58(self, manager):
        """test_request_granted."""
        record, _ = await manager.prepare_request(TEST_CONN_ID)
        grant = MediationGrant(endpoint=TEST_ENDPOINT, routing_keys=[TEST_BASE58_VERKEY])
        await manager.request_granted(record, grant)
        assert record.state == MediationRecord.STATE_GRANTED
        assert record.endpoint == TEST_ENDPOINT
        assert record.routing_keys == [TEST_VERKEY]

    async def test_request_granted_did_key(self, manager):
        """test_request_granted."""
        record, _ = await manager.prepare_request(TEST_CONN_ID)
        grant = MediationGrant(endpoint=TEST_ENDPOINT, routing_keys=[TEST_VERKEY])
        await manager.request_granted(record, grant)
        assert record.state == MediationRecord.STATE_GRANTED
        assert record.endpoint == TEST_ENDPOINT
        assert record.routing_keys == [TEST_VERKEY]

    async def test_request_denied(self, manager):
        """test_request_denied."""
        record, _ = await manager.prepare_request(TEST_CONN_ID)
        deny = MediationDeny()
        await manager.request_denied(record, deny)
        assert record.state == MediationRecord.STATE_DENIED

    @pytest.mark.skip(reason="Mediation terms are not well defined in RFC 0211")
    async def test_request_denied_counter_terms(self):
        """test_request_denied_counter_terms."""

    async def test_prepare_keylist_query(self, manager):
        """test_prepare_keylist_query."""
        query = await manager.prepare_keylist_query()
        assert query.paginate.limit == -1
        assert query.paginate.offset == 0

    async def test_prepare_keylist_query_pagination(self, manager):
        """test_prepare_keylist_query_pagination."""
        query = await manager.prepare_keylist_query(paginate_limit=10, paginate_offset=20)
        assert query.paginate.limit == 10
        assert query.paginate.offset == 20

    @pytest.mark.skip(reason="Filtering is not well defined in RFC 0211")
    async def test_prepare_keylist_query_filter(self):
        """test_prepare_keylist_query_filter."""

    async def test_add_key_no_message(self, manager):
        """test_add_key_no_message."""
        update = await manager.add_key(TEST_VERKEY)
        assert update.updates
        assert update.updates[0].action == KeylistUpdateRule.RULE_ADD

    async def test_add_key_accumulate_in_message(self, manager):
        """test_add_key_accumulate_in_message."""
        update = await manager.add_key(TEST_VERKEY)
        await manager.add_key(recipient_key=TEST_ROUTE_VERKEY, message=update)
        assert update.updates
        assert len(update.updates) == 2
        assert update.updates[0].action == KeylistUpdateRule.RULE_ADD
        assert update.updates[1].action == KeylistUpdateRule.RULE_ADD
        assert update.updates[0].recipient_key == TEST_VERKEY
        assert update.updates[1].recipient_key == TEST_ROUTE_VERKEY

    async def test_remove_key_no_message(self, manager):
        """test_remove_key_no_message."""
        update = await manager.remove_key(TEST_VERKEY)
        assert update.updates
        assert update.updates[0].action == KeylistUpdateRule.RULE_REMOVE

    async def test_remove_key_accumulate_in_message(self, manager):
        """test_remove_key_accumulate_in_message."""
        update = await manager.remove_key(TEST_VERKEY)
        await manager.remove_key(recipient_key=TEST_ROUTE_VERKEY, message=update)
        assert update.updates
        assert len(update.updates) == 2
        assert update.updates[0].action == KeylistUpdateRule.RULE_REMOVE
        assert update.updates[1].action == KeylistUpdateRule.RULE_REMOVE
        assert update.updates[0].recipient_key == TEST_VERKEY
        assert update.updates[1].recipient_key == TEST_ROUTE_VERKEY

    async def test_add_remove_key_mix(self, manager):
        """test_add_remove_key_mix."""
        update = await manager.add_key(TEST_VERKEY)
        await manager.remove_key(recipient_key=TEST_ROUTE_VERKEY, message=update)
        assert update.updates
        assert len(update.updates) == 2
        assert update.updates[0].action == KeylistUpdateRule.RULE_ADD
        assert update.updates[1].action == KeylistUpdateRule.RULE_REMOVE
        assert update.updates[0].recipient_key == TEST_VERKEY
        assert update.updates[1].recipient_key == TEST_ROUTE_VERKEY

    async def test_store_update_results(
        self,
        session: ProfileSession,
        manager: MediationManager,
    ):
        """test_store_update_results."""
        await RouteRecord(
            role=RouteRecord.ROLE_CLIENT,
            connection_id=TEST_CONN_ID,
            recipient_key=TEST_VERKEY,
        ).save(session)
        results = [
            KeylistUpdated(
                recipient_key=TEST_ROUTE_VERKEY,
                action=KeylistUpdateRule.RULE_ADD,
                result=KeylistUpdated.RESULT_SUCCESS,
            ),
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_REMOVE,
                result=KeylistUpdated.RESULT_SUCCESS,
            ),
        ]
        await manager.store_update_results(TEST_CONN_ID, results)
        routes = await RouteRecord.query(session)

        assert len(routes) == 1
        assert routes[0].recipient_key == TEST_ROUTE_VERKEY

        results = [
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_REMOVE,
                result=KeylistUpdated.RESULT_SUCCESS,
            ),
        ]

        with (
            mock.patch.object(
                RouteRecord, "query", mock.CoroutineMock()
            ) as mock_route_rec_query,
            mock.patch.object(
                test_module.LOGGER, "error", mock.MagicMock()
            ) as mock_logger_error,
        ):
            mock_route_rec_query.side_effect = StorageNotFoundError("no record")

            await manager.store_update_results(TEST_CONN_ID, results)
            mock_logger_error.assert_called_once()

        with (
            mock.patch.object(
                RouteRecord, "query", mock.CoroutineMock()
            ) as mock_route_rec_query,
            mock.patch.object(
                test_module.LOGGER, "error", mock.MagicMock()
            ) as mock_logger_error,
        ):
            mock_route_rec_query.return_value = [
                mock.MagicMock(delete_record=mock.CoroutineMock())
            ] * 2

            await manager.store_update_results(TEST_CONN_ID, results)
            mock_logger_error.assert_called_once()

    async def test_store_update_results_exists_relay(self, session, manager):
        """test_store_update_results_record_exists_relay."""
        await RouteRecord(
            role=RouteRecord.ROLE_CLIENT,
            recipient_key=TEST_VERKEY,
            wallet_id="test_wallet",
        ).save(session)
        results = [
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_ADD,
                result=KeylistUpdated.RESULT_SUCCESS,
            )
        ]
        await manager.store_update_results(TEST_CONN_ID, results)
        routes = await RouteRecord.query(session)

        assert len(routes) == 1
        route = routes[0]
        assert route.recipient_key == TEST_VERKEY
        assert route.wallet_id == "test_wallet"
        assert route.connection_id == TEST_CONN_ID

    async def test_store_update_results_errors(self, manager):
        """test_store_update_results with errors."""
        results = [
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_ADD,
                result=KeylistUpdated.RESULT_NO_CHANGE,
            ),
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_REMOVE,
                result=KeylistUpdated.RESULT_SERVER_ERROR,
            ),
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_REMOVE,
                result=KeylistUpdated.RESULT_CLIENT_ERROR,
            ),
        ]
        with mock.patch.object(test_module, "LOGGER", autospec=True) as mock_logger:
            await manager.store_update_results(TEST_CONN_ID, results)
        assert mock_logger.warning.call_count == 3

    async def test_get_my_keylist(self, session, manager):
        """test_get_my_keylist."""
        await RouteRecord(
            role=RouteRecord.ROLE_CLIENT,
            connection_id=TEST_CONN_ID,
            recipient_key=TEST_VERKEY,
        ).save(session)
        # Non-client record to verify filtering
        await RouteRecord(
            role=RouteRecord.ROLE_SERVER,
            connection_id=TEST_CONN_ID,
            recipient_key=TEST_ROUTE_VERKEY,
        ).save(session)
        keylist = await manager.get_my_keylist(TEST_CONN_ID)
        assert keylist
        assert len(keylist) == 1
        assert keylist[0].connection_id == TEST_CONN_ID
        assert keylist[0].recipient_key == TEST_VERKEY
        assert keylist[0].role == RouteRecord.ROLE_CLIENT



"""Wallet record."""

from typing import Any, Optional, Sequence

from marshmallow import fields, validate
from marshmallow.utils import EXCLUDE

from ...messaging.models.base_record import BaseRecord, BaseRecordSchema
from ...messaging.valid import UUID4_EXAMPLE
from ..error import WalletSettingsError


class WalletRecord(BaseRecord):
    """Represents a wallet record."""

    class Meta:
        """WalletRecord metadata."""

        schema_class = "WalletRecordSchema"

    RECORD_TYPE = "wallet_record"
    RECORD_ID_NAME = "wallet_id"

    TAG_NAMES = {"wallet_name"}

    MODE_MANAGED = "managed"
    MODE_UNMANAGED = "unmanaged"

    def __init__(
        self,
        *,
        wallet_id: Optional[str] = None,
        key_management_mode: Optional[str] = None,
        settings: Optional[dict] = None,
        # MTODO: how to make this a tag without making it
        # a constructor param
        wallet_name: Optional[str] = None,
        jwt_iat: Optional[int] = None,
        **kwargs,
    ):
        """Initialize a new WalletRecord."""
        super().__init__(wallet_id, **kwargs)
        self.key_management_mode = key_management_mode
        self.jwt_iat = jwt_iat
        self._settings = settings

    @property
    def wallet_id(self) -> str:
        """Accessor for the ID associated with this record."""
        return self._id

    @property
    def settings(self) -> dict:
        """Accessor for the context settings associated with this wallet."""
        return {**self._settings, "wallet.id": self.wallet_id}

    @property
    def wallet_name(self) -> Optional[str]:
        """Accessor for the name of the wallet."""
        return self.settings.get("wallet.name")

    @property
    def wallet_type(self) -> str:
        """Accessor for the type of the wallet."""
        return self.settings.get("wallet.type")

    @property
    def wallet_webhook_urls(self) -> Sequence[str]:
        """Accessor for webhook_urls of the wallet."""
        return self.settings.get("wallet.webhook_urls")

    @property
    def wallet_dispatch_type(self) -> str:
        """Accessor for webhook dispatch type of the wallet."""
        return self.settings.get("wallet.dispatch_type")

    @property
    def wallet_key(self) -> Optional[str]:
        """Accessor for the key of the wallet."""
        return self.settings.get("wallet.key")

    @property
    def wallet_key_derivation_method(self) -> Optional[str]:
        """Accessor for the key derivation method of the wallet."""
        return self.settings.get("wallet.key_derivation_method")

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this record."""
        return {
            prop: getattr(self, prop)
            for prop in ("settings", "key_management_mode", "jwt_iat")
        }

    @property
    def is_managed(self) -> bool:
        """Accessor to check if the key management mode is managed."""
        return self.key_management_mode == WalletRecord.MODE_MANAGED

    @property
    def requires_external_key(self) -> bool:
        """Accessor to check if the wallet requires an external key."""

        # Key not required for in_memory wallets
        if self.wallet_type == "in_memory":
            return False
        # Managed wallets have the key stored in the wallet
        elif self.is_managed:
            return False
        # All other cases the key is required
        else:
            return True

    def update_settings(self, settings: dict):
        """Update settings."""
        if "wallet.id" in settings:
            raise WalletSettingsError("wallet.id cannot be saved in settings.")
        self._settings.update(settings)

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class WalletRecordSchema(BaseRecordSchema):
    """Schema to allow serialization/deserialization of record."""

    class Meta:
        """WalletRecordSchema metadata."""

        model_class = WalletRecord
        unknown = EXCLUDE

    wallet_id = fields.Str(
        required=True,
        metadata={"description": "Wallet record ID", "example": UUID4_EXAMPLE},
    )
    key_management_mode = fields.Str(
        required=True,
        validate=validate.OneOf([WalletRecord.MODE_MANAGED, WalletRecord.MODE_UNMANAGED]),
        metadata={"description": "Mode regarding management of wallet key"},
    )
    settings = fields.Dict(
        required=False, metadata={"description": "Settings for this wallet."}
    )



"""Admin error classes."""

from ..core.error import BaseError


class AdminError(BaseError):
    """Base class for Admin-related errors."""


class AdminSetupError(AdminError):
    """Admin server setup or configuration error."""



from datetime import datetime, timedelta
from unittest import IsolatedAsyncioTestCase, mock

from ....tests.data import TEST_VC_DOCUMENT_SIGNED_ED25519
from ....tests.document_loader import custom_document_loader
from ...validation_result import PurposeResult
from ..assertion_proof_purpose import AssertionProofPurpose
from ..credential_issuance_purpose import CredentialIssuancePurpose


class TestCredentialIssuancePurpose(IsolatedAsyncioTestCase):
    async def test_properties(self):
        date = datetime.now()
        delta = timedelta(1)
        proof_purpose = CredentialIssuancePurpose(date=date, max_timestamp_delta=delta)
        proof_purpose2 = CredentialIssuancePurpose(date=date, max_timestamp_delta=delta)

        assert proof_purpose.term == "assertionMethod"
        assert proof_purpose.date == date
        assert proof_purpose.max_timestamp_delta == delta

        assert proof_purpose2 == proof_purpose
        assert proof_purpose != 10

    async def test_validate(self):
        proof_purpose = CredentialIssuancePurpose()

        with mock.patch.object(AssertionProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = PurposeResult(
                valid=True, controller={"id": TEST_VC_DOCUMENT_SIGNED_ED25519["issuer"]}
            )
            document = TEST_VC_DOCUMENT_SIGNED_ED25519.copy()
            proof = document.pop("proof")
            suite = mock.MagicMock()
            verification_method = {"controller": "controller"}

            result = proof_purpose.validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=custom_document_loader,
            )
            validate_mock.assert_called_once_with(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=custom_document_loader,
            )
            assert result.valid

    async def test_validate_x_no_issuer(self):
        proof_purpose = CredentialIssuancePurpose()

        with mock.patch.object(AssertionProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = PurposeResult(
                valid=True, controller={"id": TEST_VC_DOCUMENT_SIGNED_ED25519["issuer"]}
            )
            document = TEST_VC_DOCUMENT_SIGNED_ED25519.copy()
            document.pop("issuer")
            proof = document.pop("proof")
            suite = mock.MagicMock()
            verification_method = {"controller": "controller"}

            result = proof_purpose.validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=custom_document_loader,
            )
            assert not result.valid
            assert "issuer is required" in str(result.error)

    async def test_validate_x_no_match_issuer(self):
        proof_purpose = CredentialIssuancePurpose()

        with mock.patch.object(AssertionProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = PurposeResult(
                valid=True, controller={"id": "random_controller_id"}
            )
            document = TEST_VC_DOCUMENT_SIGNED_ED25519.copy()
            proof = document.pop("proof")
            suite = mock.MagicMock()
            verification_method = {"controller": "controller"}

            result = proof_purpose.validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=custom_document_loader,
            )
            assert not result.valid
            assert "issuer must match the verification method controller" in str(
                result.error
            )

    async def test_validate_x_super_invalid(self):
        proof_purpose = CredentialIssuancePurpose()

        with mock.patch.object(AssertionProofPurpose, "validate") as validate_mock:
            validate_mock.return_value = mock.MagicMock(valid=False)

            result = proof_purpose.validate(
                proof=mock.MagicMock(),
                document=mock.MagicMock(),
                suite=mock.MagicMock(),
                verification_method=mock.MagicMock(),
                document_loader=mock.MagicMock(),
            )

            assert not result.valid
            assert result == validate_mock.return_value



"""V2.0 issue-credential base credential format handler."""

import logging
from abc import ABC, abstractmethod
from typing import Mapping, Optional, Tuple

from .....core.error import BaseError
from .....core.profile import Profile
from .....messaging.decorators.attach_decorator import AttachDecorator
from ..messages.cred_format import V20CredFormat
from ..messages.cred_issue import V20CredIssue
from ..messages.cred_offer import V20CredOffer
from ..messages.cred_proposal import V20CredProposal
from ..messages.cred_request import V20CredRequest
from ..models.cred_ex_record import V20CredExRecord

LOGGER = logging.getLogger(__name__)

CredFormatAttachment = Tuple[V20CredFormat, AttachDecorator]


class V20CredFormatError(BaseError):
    """Credential format error under issue-credential protocol v2.0."""


class V20CredFormatHandler(ABC):
    """Base credential format handler."""

    format: V20CredFormat.Format = None

    def __init__(self, profile: Profile):
        """Initialize CredFormatHandler."""
        super().__init__()

        self._profile = profile

    @property
    def profile(self) -> Profile:
        """Accessor for the current profile instance.

        Returns:
            The profile instance for this credential format

        """
        return self._profile

    @abstractmethod
    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """

    @abstractmethod
    def get_format_data(self, message_type: str, data: dict) -> CredFormatAttachment:
        """Get credential format and attachment objects for use in cred ex messages."""

    @classmethod
    @abstractmethod
    def validate_fields(cls, message_type: str, attachment_data: dict) -> None:
        """Validate attachment data for specific message type and format."""

    @abstractmethod
    async def create_proposal(
        self, cred_ex_record: V20CredExRecord, proposal_data: Mapping
    ) -> CredFormatAttachment:
        """Create format specific credential proposal attachment data."""

    @abstractmethod
    async def receive_proposal(
        self, cred_ex_record: V20CredExRecord, cred_proposal_message: V20CredProposal
    ) -> None:
        """Receive format specific credential proposal message."""

    @abstractmethod
    async def create_offer(
        self, cred_proposal_message: V20CredProposal
    ) -> CredFormatAttachment:
        """Create format specific credential offer attachment data."""

    @abstractmethod
    async def receive_offer(
        self, cred_ex_record: V20CredExRecord, cred_offer_message: V20CredOffer
    ) -> None:
        """Receive foramt specific credential offer message."""

    @abstractmethod
    async def create_request(
        self, cred_ex_record: V20CredExRecord, request_data: Optional[Mapping] = None
    ) -> CredFormatAttachment:
        """Create format specific credential request attachment data."""

    @abstractmethod
    async def receive_request(
        self, cred_ex_record: V20CredExRecord, cred_request_message: V20CredRequest
    ) -> None:
        """Receive format specific credential request message."""

    def can_receive_request_without_offer(
        self,
    ) -> bool:
        """Can this handler receive credential request without an offer?"""
        return False

    @abstractmethod
    async def issue_credential(
        self, cred_ex_record: V20CredExRecord, retries: int = 5
    ) -> CredFormatAttachment:
        """Create format specific issue credential attachment data."""

    @abstractmethod
    async def receive_credential(
        self, cred_ex_record: V20CredExRecord, cred_issue_message: V20CredIssue
    ) -> None:
        """Create format specific issue credential message."""

    @abstractmethod
    async def store_credential(
        self, cred_ex_record: V20CredExRecord, cred_id: Optional[str] = None
    ) -> None:
        """Store format specific credential from issue credential message."""



"""Inner structure of KeylistQuery. Represents KeylistQuery.paginate."""

from marshmallow import fields

from ......messaging.models.base import BaseModel, BaseModelSchema


class KeylistQueryPaginate(BaseModel):
    """Class representing a keylist query pagination."""

    class Meta:
        """Keylist query pagination metadata."""

        schema_class = "KeylistQueryPaginateSchema"

    def __init__(self, limit: int, offset: int, **kwargs):
        """Initialize keylist query pagination object.

        Args:
            limit: limit for response count
            offset: offset value
            kwargs: additional keyword arguments for BaseModel

        """
        super().__init__(**kwargs)
        self.limit = limit
        self.offset = offset


class KeylistQueryPaginateSchema(BaseModelSchema):
    """Keylist query pagination schema."""

    class Meta:
        """Keylist query pagination schema metadata."""

        model_class = KeylistQueryPaginate

    limit = fields.Int(
        required=False,
        metadata={"description": "Limit for keylist query", "example": 30},
    )
    offset = fields.Int(
        required=False, metadata={"description": "Offset value for query", "example": 0}
    )
